# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import time

import functools
from retryz import retry

from storops.exception import VNXDiskUsedError, raise_if_err, \
    VNXSetArrayNameError
from storops.lib.common import daemon, instance_cache, clear_instance_cache
from storops.lib.resource import ResourceList, ResourceListCollection
from storops.vnx.resource.host import VNXHost
from storops.vnx.resource.metric import VNXStats
from storops.vnx.resource.nfs_share import VNXNfsShare
from storops.vnx.resource.fs_snap import VNXFsSnap
from storops.vnx.resource.mover import VNXMover
from storops.vnx.resource.vdm import VNXVdm
from storops.vnx.resource.cifs_share import VNXCifsShare
from storops.vnx.resource.cifs_server import VNXCifsServer
from storops.vnx.resource.nas_pool import VNXNasPool
from storops.vnx.nas_client import VNXNasClient
from storops.vnx.resource.fs import VNXFileSystem
from storops.vnx.resource.mirror_view import VNXMirrorView, VNXMirrorGroup, \
    VNXMirrorViewAsync, VNXMirrorGroupAsync
from storops.vnx.enums import VNXPortType, VNXPoolRaidType, VNXSPEnum, \
    VNXCtrlMethod
from storops.vnx.block_cli import CliClient
from storops.vnx.resource.block_pool import VNXPool, VNXPoolFeature
from storops.vnx.resource.cg import VNXConsistencyGroup
from storops.vnx.resource.disk import VNXDisk, VNXDiskList
from storops.vnx.resource.security import VNXBlockUser
from storops.vnx.resource.vnx_domain import VNXDomainMemberList, \
    VNXNetworkAdmin, VNXDomainNodeList, VNXStorageProcessor, \
    VNXStorageProcessorList
from storops.vnx.resource.lun import VNXLun
from storops.vnx.resource.migration import VNXMigrationSession
from storops.vnx.resource.ndu import VNXNdu, VNXNduList
from storops.vnx.resource.nqm import VNXIOClass, VNXIOPolicy
from storops.vnx.resource.port import VNXConnectionPort, VNXSPPort
from storops.vnx.resource import VNXCliResource
from storops.vnx.resource.rg import VNXRaidGroup
from storops.vnx.resource.sg import VNXStorageGroup
from storops.vnx.resource.snap import VNXSnap
from storops.vnx.resource.capacity import VNXCapacity

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class VNXSystem(VNXCliResource):
    """ The system class for VNX

    This class act as the entry point for VNX system.
    Various VNX resources could be retrieved with methods/properties
    of this class.
    """

    def __init__(self, ip=None,
                 username=None, password=None, scope=0, sec_file=None,
                 timeout=None,
                 heartbeat_interval=None,
                 naviseccli=None,
                 file_username=None, file_password=None):
        """ initialize a `VNXSystem` instance

        The `VNXSystem` instance act as a entry point for all
        VNX resources.

        :param ip: ip of one block sp
        :param username: username for block system
        :param password: password for the specified block user
        :param scope: scope of the specified block user
        :param sec_file: security file used by naviseccli
        :param timeout: naviseccli command timeout
        :param heartbeat_interval: heartbeat interval used to check the
        alive of sp.  Set to 0 if heart beat is not required.
        :param naviseccli: binary location of naviseccli in your host.
        :param file_username: username for control station login, default to
        username
        :param file_password: password for control station login, default to
        password
        :return: vnx system instance
        """
        super(VNXSystem, self).__init__()
        self._ip = ip
        self._username = username
        self._password = password
        self._scope = scope
        self._sec_file = sec_file
        self._timeout = timeout
        self._hb_interval = heartbeat_interval
        self._naviseccli = naviseccli

        self._file_username = file_username
        self._file_password = file_password

        self._cli = self._init_block_cli()

        if heartbeat_interval:
            daemon(self.update_nodes_ip)

    def _init_block_cli(self):
        return CliClient(
            self._ip,
            self._username, self._password, self._scope, self._sec_file,
            self._timeout, heartbeat_interval=self._hb_interval,
            naviseccli=self._naviseccli)

    def _init_file_cli(self):
        return VNXNasClient(self.control_station_ip,
                            self._file_username,
                            self._file_password)

    def __getstate__(self):
        d = {'ip': self._ip, 'username': self._username,
             'password': self._password, 'scope': self._scope,
             'sec_file': self._sec_file, 'naviseccli': self._naviseccli}
        return d

    def __setstate__(self, state):
        self.__init__(heartbeat_interval=0, **state)

    def __getinitargs__(self):
        # Seems not used by pickle
        return (self._ip, self._username, self._password, self._scope,
                self._sec_file, None, None, self._naviseccli)

    @property
    @instance_cache
    def _file_cli(self):
        return self._init_file_cli()

    @property
    @instance_cache
    def _ndu_list(self):
        ret = VNXNduList(self._cli)
        ret.with_no_poll()
        return ret

    @property
    @instance_cache
    def name(self):
        inst = VNXArrayName(self._cli)
        return inst.name

    @name.setter
    @clear_instance_cache
    def name(self, new_name):
        inst = VNXArrayName(self._cli)
        inst.set_name(new_name)

    def set_naviseccli(self, cli_binary):
        self._cli.set_binary(cli_binary)

    def set_block_credential(self,
                             username=None, password=None, scope=None,
                             sec_file=None):
        self._cli.set_credential(username, password, scope, sec_file)

    def update_nodes_ip(self):
        # do not use the `control_station_ip` property to avoid self loop.
        self._cli.set_ip(self.spa_ip, self.spb_ip, self._get_cs_ip())

    @clear_instance_cache
    def update(self, data=None):
        super(VNXSystem, self).update(data)
        self.update_nodes_ip()

    @property
    def heartbeat(self):
        return self._cli.heartbeat

    @property
    @instance_cache
    def spa_ip(self):
        return VNXNetworkAdmin.get_spa_ip(self._cli)

    @property
    @instance_cache
    def spa(self):
        return VNXStorageProcessor(self._cli, VNXSPEnum.SP_A, self.spa_ip)

    @property
    @instance_cache
    def spb(self):
        return VNXStorageProcessor(self._cli, VNXSPEnum.SP_B, self.spb_ip)

    def get_sp(self):
        return VNXStorageProcessorList(
            VNXStorageProcessor(self._cli, VNXSPEnum.SP_A, self.spa_ip),
            VNXStorageProcessor(self._cli, VNXSPEnum.SP_B, self.spb_ip))

    @property
    @instance_cache
    def spb_ip(self):
        return VNXNetworkAdmin.get_spb_ip(self._cli)

    @property
    def alive_sp_ip(self):
        return self.heartbeat.get_alive_sp_ip()

    @property
    @instance_cache
    def control_station_ip(self):
        return self._get_cs_ip()

    def _get_cs_ip(self):
        return VNXDomainNodeList.get_cs_ip(self.serial, self._cli)

    @property
    def domain(self):
        return VNXDomainMemberList(self._cli)

    def _get_raw_resource(self):
        return self._cli.get_agent(poll=self.poll)

    def get_pool_feature(self, poll=False):
        feature = VNXPoolFeature(self._cli)
        return self._update_poll(feature, poll)

    def get_pool(self, name=None, pool_id=None, system_lun_list=None):
        return VNXPool.get(pool_id=pool_id, name=name, cli=self._cli,
                           system_lun_list=system_lun_list)

    def get_lun(self, lun_id=None, name=None, lun_type=None):
        return VNXLun.get(self._cli, lun_id=lun_id, name=name,
                          lun_type=lun_type)

    def get_cg(self, name=None):
        return VNXConsistencyGroup.get(self._cli, name)

    def get_sg(self, name=None, system_lun_list=None):
        return VNXStorageGroup.get(self._cli, name,
                                   system_lun_list=system_lun_list)

    def get_snap(self, name=None):
        return VNXSnap.get(self._cli, name)

    def get_migration_session(self, src_lun=None):
        return VNXMigrationSession.get(self._cli, src_lun)

    def get_ndu(self, name=None):
        return VNXNdu.get(self._cli, name)

    def get_connection_port(self, sp=None, port_id=None, vport_id=None):
        return VNXConnectionPort.get(self._cli, sp, port_id, vport_id)

    @staticmethod
    def _update_poll(obj, poll):
        if obj is not None:
            obj.poll = poll
        return obj

    def get_sp_port(self, sp=None, port_id=None, poll=False):
        ret = VNXSPPort.get(self._cli, sp, port_id)
        return self._update_poll(ret, poll)

    def get_fc_port(self, sp=None, port_id=None, poll=False):
        ret = VNXSPPort.get(self._cli,
                            sp=sp,
                            port_id=port_id,
                            port_type=VNXPortType.FC)
        return self._update_poll(ret, poll)

    def get_iscsi_port(self, sp=None, port_id=None, vport_id=None,
                       has_ip=None, poll=False):
        ret = VNXConnectionPort.get(self._cli,
                                    sp=sp,
                                    port_id=port_id,
                                    vport_id=vport_id,
                                    port_type=VNXPortType.ISCSI,
                                    has_ip=has_ip)
        return self._update_poll(ret, poll)

    def get_fcoe_port(self, sp=None, port_id=None, vport_id=None, poll=False):
        ret = VNXConnectionPort.get(self._cli,
                                    sp=sp,
                                    port_id=port_id,
                                    vport_id=vport_id,
                                    port_type=VNXPortType.FCOE)

        return self._update_poll(ret, poll)

    def delete_snap(self, name):
        self._delete_resource(VNXSnap(name, self._cli))

    def create_sg(self, name):
        return VNXStorageGroup.create(name, self._cli)

    def delete_sg(self, name):
        self._delete_resource(VNXStorageGroup(name, self._cli))

    def create_cg(self, name, members=None):
        return VNXConsistencyGroup.create(self._cli, name=name,
                                          members=members)

    def delete_cg(self, name):
        self._delete_resource(VNXConsistencyGroup(name, self._cli))

    def delete_lun(self, name):
        self._delete_resource(VNXLun(name=name, cli=self._cli))

    def get_disk(self, disk_index=None):
        return VNXDisk.get(self._cli, disk_index)

    def get_available_disks(self):
        pool_feature = VNXPoolFeature(self._cli)
        pool_feature.poll = self.poll
        disks = pool_feature.available_disks
        if disks:
            disks.poll = self.poll
        else:
            disks = VNXDiskList(cli=self._cli, disk_indices='N/A')
        return disks

    def delete_disk(self, disk_index):
        self._delete_resource(VNXDisk(disk_index, self._cli))

    def install_disk(self, disk_index):
        disk = VNXDisk(disk_index, self._cli)
        disk.install()

    def get_rg(self, rg_id=None):
        return VNXRaidGroup.get(self._cli, rg_id)

    def create_rg(self, rg_id, disks, raid_type=None):
        return VNXRaidGroup.create(self._cli, rg_id, disks, raid_type)

    def delete_rg(self, rg_id):
        self._delete_resource(VNXRaidGroup(rg_id, self._cli))

    def create_pool(self, name, disks=None, raid_type=None):
        @retry(on_error=VNXDiskUsedError)
        def create_with_default_disks():
            _disks = self.get_available_disks()
            disk_count = VNXPoolRaidType.parse(raid_type).min_disk_requirement
            _disks.same_disks(disk_count)
            return VNXPool.create(self._cli, name, _disks, raid_type)

        if raid_type is None:
            raid_type = VNXPoolRaidType.RAID5
        if disks is None:
            ret = create_with_default_disks()
        else:
            ret = VNXPool.create(self._cli, name, disks, raid_type)

        return ret

    def _delete_resource(self, resource):
        resource.poll = self.poll
        resource.delete()

    def delete_pool(self, name=None, pool_id=None):
        self._delete_resource(VNXPool(pool_id, name, self._cli))

    def stop_heart_beat(self):
        self._cli.heartbeat.stop()

    def is_compression_enabled(self):
        return self._ndu_list.is_compression_enabled()

    def is_dedup_enabled(self):
        return self._ndu_list.is_dedup_enabled()

    def is_snap_enabled(self):
        return self._ndu_list.is_snap_enabled()

    def is_mirror_view_async_enabled(self):
        return self._ndu_list.is_mirror_view_async_enabled()

    def is_mirror_view_sync_enabled(self):
        return self._ndu_list.is_mirror_view_sync_enabled()

    def is_mirror_view_enabled(self):
        return self._ndu_list.is_mirror_view_enabled()

    def is_thin_enabled(self):
        return self._ndu_list.is_thin_enabled()

    def is_sancopy_enabled(self):
        return self._ndu_list.is_sancopy_enabled()

    def is_auto_tiering_enabled(self):
        return self._ndu_list.is_auto_tiering_enabled()

    def is_fast_cache_enabled(self):
        return self._ndu_list.is_fast_cache_enabled()

    def delete_hba(self, hba_uid):
        return VNXSPPort.delete_hba(self._cli, hba_uid)

    def get_block_user(self, name=None):
        return VNXBlockUser.get(cli=self._cli, name=name)

    def create_block_user(self, name, password, scope=None, role=None):
        return VNXBlockUser.create(self._cli, name, password, scope, role)

    def get_mirror_view(self, name=None):
        return VNXMirrorView.get(self._cli, name)

    def create_mirror_view(self, name, src_lun, use_write_intent_log=True):
        return VNXMirrorView.create(self._cli, name, src_lun,
                                    use_write_intent_log)

    def get_mirror_group(self, name=None):
        return VNXMirrorGroup.get(self._cli, name)

    def create_mirror_group(self, name, mirror=None):
        return VNXMirrorGroup.create(self._cli, name, mirror)

    def get_mirror_view_async(self, name=None):
        return VNXMirrorViewAsync.get(self._cli, name)

    def create_mirror_view_async(self, name, src_lun):
        return VNXMirrorViewAsync.create(self._cli, name, src_lun)

    def get_mirror_group_async(self, name=None):
        return VNXMirrorGroupAsync.get(self._cli, name)

    def create_mirror_group_async(self, name, mirror=None):
        return VNXMirrorGroupAsync.create(self._cli, name, mirror)

    def get_host(self, name=None):
        return VNXHost.get(self._cli, name)

    def get_capacity(self):
        return VNXCapacity.get(cli=self._cli)

    def create_ioclass(self, name, iotype, luns=None,
                       ctrlmethod=VNXCtrlMethod.NO_CTRL):
        return VNXIOClass.create(cli=self._cli, name=name, iotype=iotype,
                                 luns=luns, ctrlmethod=ctrlmethod)

    def get_ioclass(self, name=None):
        return VNXIOClass.get(cli=self._cli, name=name)

    def create_policy(self, name, ioclasses=None, fail_action=None,
                      time_limit=None, eval_window=None):
        return VNXIOPolicy.create(
            cli=self._cli, name=name, ioclasses=ioclasses,
            fail_action=fail_action, time_limit=time_limit,
            eval_window=eval_window)

    def get_policy(self, name=None):
        return VNXIOPolicy.get(cli=self._cli, name=name)

    def stop_policy(self):
        return VNXIOPolicy.stop_policy(cli=self._cli)

    def get_file_system(self, name=None, fs_id=None):
        return VNXFileSystem.get(cli=self._file_cli, name=name, fs_id=fs_id)

    def get_nas_pool(self, name=None, pool_id=None):
        return VNXNasPool.get(cli=self._file_cli, name=name, pool_id=pool_id)

    def get_cifs_server(self, name=None, mover_id=None, is_vdm=None):
        return VNXCifsServer.get(cli=self._file_cli, name=name,
                                 mover_id=mover_id, is_vdm=is_vdm)

    def create_cifs_server(self, name, mover_id=None, is_vdm=False,
                           workgroup=None, domain=None,
                           interfaces=None, alias_name=None,
                           local_admin_password=None):
        return VNXCifsServer.create(
            cli=self._file_cli, name=name, mover_id=mover_id, is_vdm=is_vdm,
            workgroup=workgroup, domain=domain, interfaces=interfaces,
            alias_name=alias_name, local_admin_password=local_admin_password)

    def get_cifs_share(self, name=None, mover=None, server_name=None):
        return VNXCifsShare.get(
            self._file_cli, name=name, mover=mover, server_name=server_name)

    def get_file_system_snap(self, name=None, snap_id=None):
        return VNXFsSnap.get(cli=self._file_cli, name=name, snap_id=snap_id)

    def get_mover(self, name=None, mover_id=None, is_vdm=False):
        if is_vdm:
            ret = VNXVdm.get(cli=self._file_cli, name=name, vdm_id=mover_id)
        else:
            ret = VNXMover.get(cli=self._file_cli, name=name,
                               mover_id=mover_id)
        return ret

    def get_nfs_share(self, mover=None, path=None):
        return VNXNfsShare.get(cli=self._file_cli, mover=mover, path=path)

    def _default_rsc_list_with_perf_stats(self):
        sps = self.get_sp()
        lun_list = self.get_lun()
        disks = self.get_disk()
        sp_ports = self.get_sp_port()
        pools = self.get_pool(system_lun_list=lun_list)
        sgs = self.get_sg(system_lun_list=lun_list)
        ret = (sps, lun_list, disks, sp_ports, pools, sgs)

        for resource_list in ret:
            resource_list.with_no_poll()
        return ret

    def get_rsc_list_2(self, rsc_clz_list=None):
        """get the list of resource list to collect based on clz list

        :param rsc_clz_list: the list of classes to collect
        :return: filtered list of resource list,
                 like [VNXLunList(), VNXDiskList()]
        """
        rsc_list_2 = self._default_rsc_list_with_perf_stats()
        if rsc_clz_list is None:
            rsc_clz_list = ResourceList.get_rsc_clz_list(rsc_list_2)

        return [rsc_list
                for rsc_list in rsc_list_2
                if rsc_list.get_resource_class() in rsc_clz_list]

    def collect_perf_record(self, clz_list):
        log.info('start collecting counters of vnx {}.'.format(self._ip))
        start = time.time()
        rsc_list_2 = self.get_rsc_list_2(clz_list)
        record = ResourceListCollection(rsc_list_2)
        record.update()
        log.info('end collecting counters of vnx {}.  collection took '
                 '{:.3f} seconds.'.format(self._ip, time.time() - start))
        return record

    def enable_perf_stats(self, rsc_clz_list=None):
        VNXStats.get(self._cli).enable_stats()
        f = functools.partial(self.collect_perf_record, clz_list=rsc_clz_list)
        self._cli.enable_perf_metric(60, f, rsc_clz_list)
        return self.get_rsc_list_2(rsc_clz_list)

    def disable_perf_stats(self, disable_counter_collection=False):
        if disable_counter_collection:
            VNXStats.get(self._cli).disable_stats()
        self._cli.disable_perf_metric()

    def is_perf_stats_enabled(self):
        return self._cli.is_perf_metric_enabled()

    def is_counter_collection_enabled(self):
        return VNXStats.get(self._cli).is_enabled()

    def enable_persist_perf_stats(self):
        rsc_list = self._default_rsc_list_with_perf_stats()
        self._cli.persist_perf_stats(rsc_list)

    def disable_persist_perf_stats(self):
        self._cli.persist_perf_stats(None)

    def is_perf_stats_persisted(self):
        return self._cli.is_perf_stats_persisted()

    def __del__(self):
        del self._cli


class VNXArrayName(VNXCliResource):
    def __init__(self, cli=None):
        super(VNXArrayName, self).__init__()
        self._cli = cli

    def _get_raw_resource(self):
        return self._cli.get_array_name()

    def set_name(self, new_name):
        out = self._cli.set_array_name(new_name)
        raise_if_err(out, default=VNXSetArrayNameError)


class VNXAgent(VNXCliResource):
    def __init__(self, cli=None):
        super(VNXAgent, self).__init__()
        self._cli = cli

    def _get_raw_resource(self):
        return self._cli.get_agent(poll=self.poll)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

from unittest import TestCase

import ddt
from hamcrest import equal_to, assert_that, instance_of, none, raises, calling

from storops.exception import UnityResourceNotFoundError, \
    UnityFileSystemNameAlreadyExisted, UnitySnapNameInUseError, \
    UnityFileSystemSizeTooSmallError, UnityShareShrinkSizeTooLargeError, \
    UnityShareShrinkSizeTooSmallError, UnityLocalReplicationFsNameNotSameError
from storops.unity.enums import FilesystemTypeEnum, TieringPolicyEnum, \
    FSSupportedProtocolEnum, AccessPolicyEnum, FSFormatEnum, \
    ResourcePoolFullPolicyEnum, HostIOSizeEnum, NFSShareDefaultAccessEnum, \
    JobStateEnum
from storops.unity.resource.cifs_share import UnityCifsShareList
from storops.unity.resource.filesystem import UnityFileSystem, \
    UnityFileSystemList
from storops.unity.resource.nas_server import UnityNasServer
from storops.unity.resource.pool import UnityPool
from storops.unity.resource.remote_system import UnityRemoteSystem
from storops.unity.resource.storage_resource import UnityStorageResource
from storops.unity.resource.health import UnityHealth
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


@ddt.ddt
class UnityFileSystemTest(TestCase):
    @patch_rest
    def test_properties(self):
        fs = UnityFileSystem(_id='fs_2', cli=t_rest())
        assert_that(fs.id, equal_to('fs_2'))
        assert_that(fs.type, equal_to(FilesystemTypeEnum.FILESYSTEM))
        assert_that(fs.tiering_policy,
                    equal_to(TieringPolicyEnum.AUTOTIER_HIGH))
        assert_that(fs.supported_protocols,
                    equal_to(FSSupportedProtocolEnum.CIFS))
        assert_that(fs.access_policy, equal_to(AccessPolicyEnum.WINDOWS))
        assert_that(fs.format, equal_to(FSFormatEnum.UFS64))
        assert_that(fs.host_io_size, equal_to(HostIOSizeEnum.GENERAL_8K))
        assert_that(fs.pool_full_policy,
                    equal_to(ResourcePoolFullPolicyEnum.DELETE_ALL_SNAPS))
        assert_that(fs.health, instance_of(UnityHealth))
        assert_that(fs.name, equal_to('esa_cifs1'))
        assert_that(fs.description, equal_to(''))
        assert_that(fs.size_total, equal_to(5368709120))
        assert_that(fs.size_used, equal_to(1642971136))
        assert_that(fs.size_allocated, equal_to(3221225472))
        assert_that(fs.is_read_only, equal_to(False))
        assert_that(fs.is_thin_enabled, equal_to(True))
        assert_that(fs.is_cifs_sync_writes_enabled, equal_to(False))
        assert_that(fs.is_cifs_op_locks_enabled, equal_to(True))
        assert_that(fs.is_cifs_notify_on_write_enabled, equal_to(False))
        assert_that(fs.is_cifs_notify_on_access_enabled, equal_to(False))
        assert_that(fs.cifs_notify_on_change_dir_depth, equal_to(512))
        assert_that(fs.metadata_size, equal_to(3489660928))
        assert_that(fs.metadata_size_allocated, equal_to(3221225472))
        assert_that(fs.per_tier_size_used, equal_to([6442450944, 0, 0]))
        assert_that(fs.snaps_size, equal_to(0))
        assert_that(fs.snaps_size_allocated, equal_to(0))
        assert_that(fs.snap_count, equal_to(0))
        assert_that(fs.is_smbca, equal_to(False))
        assert_that(fs.storage_resource, instance_of(UnityStorageResource))
        assert_that(fs.pool, instance_of(UnityPool))
        assert_that(fs.nas_server, instance_of(UnityNasServer))
        assert_that(fs.cifs_share, instance_of(UnityCifsShareList))
        assert_that(len(fs.cifs_share), equal_to(1))
        assert_that(fs.nfs_share, none())

    @patch_rest
    def test_get_all(self):
        fs_list = UnityFileSystemList(cli=t_rest())
        assert_that(len(fs_list), equal_to(3))

    @patch_rest
    def test_delete_fs_9(self):
        fs = UnityFileSystem(_id='fs_9', cli=t_rest())
        resp = fs.delete(force_snap_delete=True, force_vvol_delete=True)
        assert_that(resp.is_ok(), equal_to(True))
        assert_that(resp.job.existed, equal_to(False))

    @patch_rest
    def test_delete_not_found(self):
        def f():
            fs = UnityFileSystem(_id='fs_99', cli=t_rest())
            fs.delete(force_snap_delete=True, force_vvol_delete=True)

        assert_that(f, raises(UnityResourceNotFoundError))

    @patch_rest
    def test_extend(self):
        fs = UnityFileSystem(_id='fs_8', cli=t_rest())
        resp = fs.extend(1024 ** 3 * 5)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_extend_size_error(self):
        def f():
            fs = UnityFileSystem(_id='fs_8', cli=t_rest())
            fs.extend(1024 ** 3 * 2)

        assert_that(f, raises(UnityFileSystemSizeTooSmallError,
                              'size is too small'))

    @patch_rest
    def test_shrink_success(self):
        fs = UnityFileSystem(_id='fs_8', cli=t_rest())
        resp = fs.shrink(1024 ** 3 * 2.5)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_shrink_size_too_large_error(self):
        def f():
            fs = UnityFileSystem(_id='fs_8', cli=t_rest())
            fs.storage_resource.size_total = 1024 ** 3 * 3
            fs.shrink(1024 ** 3 * 4)

        message = 'Reject shrink share request, ' \
                  'the new size should be smaller than original.'
        assert_that(f, raises(UnityShareShrinkSizeTooLargeError, message))

    @patch_rest
    def test_shrink_size_too_small_error(self):
        def f():
            fs = UnityFileSystem(_id='fs_8', cli=t_rest())
            fs.storage_resource.size_used = 1024 ** 3 * 3
            fs.shrink(1024 ** 3 * 2)

        message = 'Reject shrink share request, ' \
                  'the new size should be larger than used.'
        assert_that(f, raises(UnityShareShrinkSizeTooSmallError, message))

    @patch_rest
    def test_create_success(self):
        fs = UnityFileSystem.create(
            t_rest(), 'pool_1', 'nas_2', 'fs3', 3 * 1024 ** 3,
            proto=FSSupportedProtocolEnum.CIFS,
            tiering_policy=TieringPolicyEnum.AUTOTIER_HIGH)
        assert_that(fs.get_id(), equal_to('fs_12'))

    @patch_rest
    def test_delete_filesystem_async(self):
        fs = UnityFileSystem(_id='fs_14', cli=t_rest())
        resp = fs.delete(async_mode=True)
        assert_that(resp.is_ok(), equal_to(True))
        job = resp.job
        assert_that(job.id, equal_to('N-345'))
        assert_that(job.state, equal_to(JobStateEnum.RUNNING))
        assert_that(job.description, equal_to(
            'job.applicationprovisioningservice.job.DeleteApplication'))
        assert_that(str(job.state_change_time),
                    equal_to('2016-03-22 10:39:20.097000+00:00'))
        assert_that(str(job.submit_time),
                    equal_to('2016-03-22 10:39:20.033000+00:00'))
        assert_that(str(job.est_remain_time), equal_to('0:00:29'))
        assert_that(job.progress_pct, equal_to(0))

    @patch_rest
    def test_create_existed(self):
        def f():
            UnityFileSystem.create(
                t_rest(), 'pool_1', 'nas_2', 'fs3', 3 * 1024 ** 3,
                proto=FSSupportedProtocolEnum.NFS)

        assert_that(f, raises(UnityFileSystemNameAlreadyExisted,
                              'file system name has already'))

    @patch_rest
    def test_create_nfs_share_success(self):
        fs = UnityFileSystem(_id='fs_9', cli=t_rest())
        share = fs.create_nfs_share(
            'ns1', share_access=NFSShareDefaultAccessEnum.READ_WRITE)
        assert_that(share.name, equal_to('ns1'))
        assert_that(share.id, equal_to('NFSShare_4'))

    @patch_rest
    def test_create_cifs_share_success(self):
        fs = UnityFileSystem(_id='fs_8', cli=t_rest())
        share = fs.create_cifs_share('cs1')
        assert_that(share.name, equal_to('cs1'))
        assert_that(share.existed, equal_to(True))

    @patch_rest
    def test_create_snap_success(self):
        fs = UnityFileSystem(_id='fs_8', cli=t_rest())
        snap = fs.create_snap()
        assert_that(snap.existed, equal_to(True))
        assert_that(snap.storage_resource, equal_to(fs.storage_resource))

    @patch_rest
    def test_create_snap_name_existed(self):
        def f():
            fs = UnityFileSystem(_id='fs_8', cli=t_rest())
            fs.create_snap(name='2016-03-15_10:56:08')

        assert_that(f, raises(UnitySnapNameInUseError, 'in use'))

    @patch_rest
    def test_create_snap_fs_snap_existed(self):
        def f():
            fs = UnityFileSystem(_id='fs_8', cli=t_rest())
            fs.create_snap('s1')

        assert_that(f, raises(UnitySnapNameInUseError, 'in use'))

    @patch_rest
    def test_fs_snapshots(self):
        fs = UnityFileSystem(_id='fs_5', cli=t_rest())
        assert_that(len(fs.snapshots), equal_to(2))

    @patch_rest
    def test_has_snap_destroying(self):
        fs = UnityFileSystem(_id='fs_5', cli=t_rest())
        assert_that(fs.has_snap(), equal_to(False))

    @patch_rest
    def test_has_snap_true(self):
        fs = UnityFileSystem(_id='fs_8', cli=t_rest())
        assert_that(fs.has_snap(), equal_to(True))

    @patch_rest
    @ddt.data(
        {'dst_fs_name': None, 'remote_system': None, 'rep_name': None,
         'dst_size': None, 'is_dst_thin': None, 'dst_tiering_policy': None,
         'is_dst_compression': None,
         'expected_rep_session_id': '171798691895_APM00192210744_0000'
                                    '_171798691923_APM00192210744_0000',
         },
        {'dst_fs_name': 'fs-liangr', 'remote_system': 'RS_6',
         'rep_name': 'fs-rep', 'dst_size': 107374182400, 'is_dst_thin': True,
         'dst_tiering_policy': TieringPolicyEnum.AUTOTIER_HIGH,
         'is_dst_compression': False,
         'expected_rep_session_id': '171798691895_APM00192210744_0000'
                                    '_171798691868_FNM00184901113_0000',
         },
    )
    @ddt.unpack
    def test_replicate_with_dst_resource_provisioning(self,
                                                      dst_fs_name,
                                                      remote_system,
                                                      rep_name,
                                                      dst_size,
                                                      is_dst_thin,
                                                      dst_tiering_policy,
                                                      is_dst_compression,
                                                      expected_rep_session_id):
        fs = UnityFileSystem.get(cli=t_rest(), _id='fs_4')
        if remote_system:
            remote_system = UnityRemoteSystem(_id=remote_system, cli=t_rest())
        rep_session = fs.replicate_with_dst_resource_provisioning(
            60, 'pool_1', dst_fs_name=dst_fs_name,
            remote_system=remote_system, replication_name=rep_name,
            dst_size=dst_size, is_dst_thin=is_dst_thin,
            dst_tiering_policy=dst_tiering_policy,
            is_dst_compression=is_dst_compression)
        assert_that(rep_session.id, equal_to(expected_rep_session_id))

    @patch_rest
    def test_replicate_with_dst_resource_provisioning_fs_name_error(self):
        fs = UnityFileSystem.get(cli=t_rest(), _id='fs_4')
        assert_that(
            calling(
                fs.replicate_with_dst_resource_provisioning
            ).with_args(
                60, 'pool_1', dst_fs_name='fs-liangr-rep-dst'),
            raises(UnityLocalReplicationFsNameNotSameError,
                   'dst_fs_name passed in for creating filesystem local '
                   'replication should be same as source filesystem name '
                   'or None')
        )



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Ryan Liang'

log = logging.getLogger(__name__)


class UnityRemoteSystem(UnityResource):

    @classmethod
    def create(cls, cli, management_address,
               local_username=None, local_password=None,
               remote_username=None, remote_password=None,
               connection_type=None):
        """
        Configures a remote system for remote replication.

        :param cls: this class.
        :param cli: the rest client.
        :param management_address: the management IP address of the remote
            system.
        :param local_username: administrative username of local system.
        :param local_password: administrative password of local system.
        :param remote_username: administrative username of remote system.
        :param remote_password: administrative password of remote system.
        :param connection_type: `ReplicationCapabilityEnum`. Replication
            connection type to the remote system.
        :return: the newly created remote system.
        """

        req_body = cli.make_body(
            managementAddress=management_address, localUsername=local_username,
            localPassword=local_password, remoteUsername=remote_username,
            remotePassword=remote_password, connectionType=connection_type)

        resp = cli.post(cls().resource_class, **req_body)
        resp.raise_if_err()
        return cls.get(cli, resp.resource_id)

    def modify(self, management_address=None, username=None, password=None,
               connection_type=None):
        """
        Modifies a remote system for remote replication.

        :param management_address: same as the one in `create` method.
        :param username: username for accessing the remote system.
        :param password: password for accessing the remote system.
        :param connection_type: same as the one in `create` method.
        """
        req_body = self._cli.make_body(
            managementAddress=management_address, username=username,
            password=password, connectionType=connection_type)

        resp = self.action('modify', **req_body)
        resp.raise_if_err()
        return resp

    def verify(self, connection_type=None):
        """
        Verifies and update the remote system settings.

        :param connection_type: same as the one in `create` method.
        """
        req_body = self._cli.make_body(connectionType=connection_type)

        resp = self.action('verify', **req_body)
        resp.raise_if_err()
        return resp


class UnityRemoteSystemList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityRemoteSystem



# coding=utf-8
# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Ryan Liang'

log = logging.getLogger(__name__)


class UnityPotentialHost(UnityResource):
    pass


class UnityPotentialHostList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityPotentialHost


class UnityHostContainer(UnityResource):
    @classmethod
    def create(cls, cli, service_type, target_name, target_address, username,
               password, description=None, local_username=None,
               local_password=None, potential_hosts=None):
        vasa_provider_parameters = None
        if local_username and local_password:
            vasa_provider_parameters = {'localUsername': local_username,
                                        'localPassword': local_password}
        resp = cli.post(cls().resource_class,
                        serviceType=service_type,
                        targetName=target_name,
                        targetAddress=target_address,
                        username=username,
                        password=password,
                        description=description,
                        vasaProviderParameters=vasa_provider_parameters,
                        potentialHosts=potential_hosts)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    @classmethod
    def recommend(cls, cli, address, username, password):
        """ Get a list of candidate hosts that can be imported for a VC or an
        ESXi server.

        :return: a tuple of (type, container_name, container_version,
            vasa_provider_state, potential_hosts), while type is of type
            `HostContainerTypeEnum` which is the type of the VC/ESXi server,
            container_name is the name of VC or ESXi server, container_version
            is the software version of VC or ESXi server,
            vasa_provider_state is of type `VasaProviderStateEnum` which
            indicates storage system’s VASA vendor provider state on the
            vCenter, potential_hosts is a list of `UnityPotentialHost` which is
            the list of candidate hosts that can be imported.
        """
        resp = cli.type_action(cls().resource_class,
                               'recommend',
                               address=address, username=username,
                               password=password)
        resp.raise_if_err()
        content = resp.first_content
        return (content.get('type'),
                content.get('containerName'),
                content.get('containerVersion'),
                content.get('vasaProviderState'),
                content.get('potentialHosts'))


class UnityHostContainerList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityHostContainer



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import json
import os
from unittest import TestCase

from hamcrest import assert_that, equal_to, only_contains, raises, none

from storops.exception import UnityNasServerNameUsedError, UnityException
from storops.unity.resp import RestResponse
from storops_test.utils import read_test_file

__author__ = 'Cedric Zhuang'


def read_json(folder, filename):
    raw = read_test_file(os.path.join('unity', 'rest_data', folder), filename)
    return json.loads(raw, encoding='utf-8')


def read_error_json(filename):
    return read_json('error', filename)


class RestResponseTest(TestCase):
    def test_has_next_page(self):
        resp = RestResponse(read_json('metric', 'metrics_page_1.json'))
        assert_that(resp.next_page, equal_to(2))
        assert_that(resp.has_next_page, equal_to(True))

    def test_has_current_page(self):
        resp = RestResponse(read_json('metric', 'metrics_page_1.json'))
        assert_that(resp.current_page, equal_to(1))


class UnityErrorTest(TestCase):
    def test_get_properties(self):
        body = read_error_json('409.json')
        resp = RestResponse(body)
        assert_that(resp.has_error(), equal_to(True))
        err = resp.error
        assert_that(err.existed, equal_to(True))
        assert_that(str(err.created),
                    equal_to('2016-03-16 15:13:45.977000+00:00'))
        assert_that(err.get_messages(), only_contains(
            'The name of NAS server is already in use. '
            '(Error Code:0x6702024)'))
        assert_that(err.http_status_code, equal_to(409))
        assert_that(err.error_code, equal_to(108011556))

    def test_raise_if_err_409(self):
        def f():
            body = read_error_json('409.json')
            resp = RestResponse(body)
            resp.raise_if_err()

        assert_that(f, raises(UnityNasServerNameUsedError, 'in use'))

    def test_raise_if_err_nothing(self):
        body = read_error_json('200.json')
        resp = RestResponse(body)
        resp.raise_if_err()


class UnityExceptionTest(TestCase):
    def test_unity_exception_error_code(self):
        resp = RestResponse(read_error_json('409.json'))
        ex = UnityException(resp.error)
        assert_that(ex.error_code, equal_to(108011556))

    def test_unity_exception_default_error_code(self):
        assert_that(UnityException().error_code, none())



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, has_items, raises, instance_of

from storops import MetricTypeEnum
from storops.exception import UnityMetricQueryNotFoundError
from storops.unity.calculator import IdValues
from storops.unity.resource.metric import UnityMetric, UnityMetricList, \
    UnityMetricRealTimeQuery, UnityMetricRealTimeQueryList
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


@patch_rest
def get_query_result(query_id, paths=None):
    return UnityMetricRealTimeQuery(
        cli=t_rest(), _id=query_id).get_query_result(paths)


qr_6 = get_query_result(6)
qr_14 = get_query_result(14)
qr_17 = get_query_result(17)
qr_34 = get_query_result(34)
qr_128 = get_query_result(128)
qr_130 = get_query_result(130)


class UnityMetricTest(TestCase):
    def verify_metric_10234(self, metric):
        assert_that(metric.id, equal_to(10234))
        assert_that(metric.is_historical_available, equal_to(False))
        assert_that(metric.is_realtime_available, equal_to(True))
        assert_that(metric.name, equal_to('Block Cache Clean Pages'))
        assert_that(metric.type, equal_to(MetricTypeEnum.FACT))
        assert_that(metric.unit_display_string, equal_to('Count'))
        assert_that(metric.path,
                    equal_to('sp.*.blockCache.global.summary.cleanPages'))

    def verify_metric_14732(self, metric):
        assert_that(metric.id, equal_to(14732))
        assert_that(metric.is_historical_available, equal_to(False))
        assert_that(metric.is_realtime_available, equal_to(True))
        assert_that(metric.name, equal_to('Total Scan'))
        assert_that(metric.type, equal_to(MetricTypeEnum.COUNTER_64))
        assert_that(metric.unit_display_string, equal_to('Requests'))
        assert_that(metric.description,
                    equal_to('Total number of antivirus requests'))
        assert_that(metric.path,
                    equal_to('sp.*.virusChecker.request.requests'))

    @patch_rest
    def test_get_properties(self):
        metric = UnityMetric(_id=10234, cli=t_rest())
        self.verify_metric_10234(metric)

    @patch_rest
    def test_get_all(self):
        metrics = UnityMetricList(cli=t_rest())
        assert_that(len(metrics), equal_to(2411))
        self.verify_metric_14732(*filter(lambda m: m.id == 14732, metrics))
        self.verify_metric_10234(*filter(lambda m: m.id == 10234, metrics))


class UnityMetricRealTimeQueryTest(TestCase):
    @patch_rest
    def test_create_query(self):
        paths = ['sp.*.physical.disk.*.reads', 'sp.*.physical.disk.*.writes']
        query = UnityMetricRealTimeQuery.create(t_rest(), 30, paths)
        assert_that(query.paths, has_items(*paths))

    @patch_rest
    def test_delete_query(self):
        def f():
            query = UnityMetricRealTimeQuery(_id=4, cli=t_rest())
            query.delete()

        assert_that(f, raises(UnityMetricQueryNotFoundError, 'ID not found'))

    @patch_rest
    def test_properties(self):
        query = UnityMetricRealTimeQuery(_id=2, cli=t_rest())
        self.verify_query_2(query)

    @patch_rest
    def test_get_all(self):
        query_list = UnityMetricRealTimeQueryList(cli=t_rest())
        assert_that(len(query_list), equal_to(4))
        self.verify_query_2(query_list[0])

    def verify_query_2(self, query):
        assert_that(query.id, equal_to(2))
        assert_that(query.interval, equal_to(300))
        assert_that(query.expiration.minute, equal_to(30))
        assert_that(query.paths,
                    has_items('sp.*.physical.disk.*.reads',
                              'sp.*.storage.lun.*.busyTime'))

    @patch_rest
    def test_sort_by_path(self):
        query_list = UnityMetricRealTimeQueryList(cli=t_rest())
        query_list.sort_by_path()
        assert_that(list(map(len, query_list.paths)),
                    equal_to([89, 88, 20, 1]))

    @patch_rest
    def test_all_filter_by_id(self):
        query_list = UnityMetricRealTimeQueryList(
            cli=t_rest(), id_list=[17, 22])
        assert_that(query_list.id, equal_to([17, 22]))

    @patch_rest
    def test_set_id_list(self):
        query_list = UnityMetricRealTimeQueryList(cli=t_rest())
        assert_that(len(query_list), equal_to(4))
        ret = query_list.set_id_list([17, 22])
        assert_that(ret, instance_of(UnityMetricRealTimeQueryList))
        assert_that(len(query_list), equal_to(2))

    @patch_rest
    def test_get_query_found_in_one(self):
        paths = ['sp.*.storage.lun.*.reads',
                 'sp.*.storage.lun.*.writes',
                 'sp.*.blockCache.global.summary.dirtyBytes']
        queries = UnityMetricRealTimeQuery.get_query_list(t_rest(), 300, paths)
        assert_that(len(queries), equal_to(1))
        assert_that(queries[0].paths, has_items(*paths))

    @patch_rest
    def test_get_query_found_in_two(self):
        paths = ['sp.*.blockCache.global.summary.dirtyBytes',
                 'sp.*.platform.storageProcessorTemperature']
        queries = UnityMetricRealTimeQuery.get_query_list(t_rest(), 300, paths)
        assert_that(len(queries), equal_to(2))

    @patch_rest
    def test_get_query_create_one(self):
        paths = ['sp.*.blockCache.global.summary.dirtyBytes',
                 'sp.*.platform.storageProcessorTemperature',
                 'sp.*.store.scsiBusDevice.*.calls']

        queries = UnityMetricRealTimeQuery.get_query_list(t_rest(), 300, paths)
        assert_that(len(queries._id_list), equal_to(3))

    @patch_rest
    def test_get_query_result_all(self):
        queries = UnityMetricRealTimeQueryList(cli=t_rest(), interval=300)
        assert_that(len(queries), equal_to(2))
        assert_that(len(queries.get_query_result()), equal_to(90))

    @patch_rest
    def test_get_query_result_filter_by_path(self):
        queries = UnityMetricRealTimeQueryList(cli=t_rest(), interval=300)
        paths = ['sp.*.blockCache.global.summary.dirtyBytes',
                 'sp.*.fibreChannel.blockSize',
                 'sp.*.storage.vvol.file.*.writes',
                 'not.found']
        assert_that(len(queries.get_query_result(paths)), equal_to(3))


class UnityMetricQueryResultTest(TestCase):
    def verify_disk_reads_value(self, result):
        assert_that(result.path, equal_to('sp.*.physical.disk.*.reads'))
        assert_that(result.query_id, equal_to(6))
        assert_that(result.timestamp.minute, equal_to(30))
        assert_that(len(result.values), equal_to(2))

        spa_values = result.values['spa']
        assert_that(len(spa_values), equal_to(26))

    @patch_rest
    def test_get_query_result_all(self):
        results = qr_6
        assert_that(len(results), equal_to(2))

        self.verify_disk_reads_value(
            results.by_path('sp.*.physical.disk.*.reads'))

    @patch_rest
    def test_get_query_result_filtered(self):
        paths = ['sp.*.blockCache.global.summary.dirtyBytes',
                 'sp.*.fibreChannel.blockSize',
                 'sp.*.nfs.v2.op.*.failures']
        results = get_query_result(3, paths)
        assert_that(len(results), equal_to(3))
        assert_that(results.path, has_items(*paths))

    @patch_rest
    def test_sum_sp(self):
        result = qr_6.by_path('sp.*.physical.disk.*.reads')
        sum_sp = result.sum_sp()
        assert_that(len(sum_sp), equal_to(26))
        assert_that(sum_sp['dpe_disk_8'], equal_to(122362))
        assert_that(sum_sp['dpe_disk_1'], equal_to(839944))

    @patch_rest
    def test_diff_seconds(self):
        assert_that(qr_14.diff_seconds(qr_6), equal_to(163800.0))

    @patch_rest
    def test_merge_result_list(self):
        r1 = get_query_result(18)
        r2 = get_query_result(19)
        assert_that(len(r1), equal_to(4))
        assert_that(len(r2), equal_to(3))
        r1.merge(r2)
        assert_that(len(r1), equal_to(6))

    @patch_rest
    def test_sp_values(self):
        result = qr_34.by_path('sp.*.cifs.smb1.basic.writes')
        assert_that(result.sp_values, instance_of(IdValues))
        assert_that(result.sp_values['spa'], equal_to(500))

    @patch_rest
    def test_sum_sp_values(self):
        result = qr_128.by_path('sp.*.cpu.summary.busyTicks')
        expected = 92496181 + 944939818
        assert_that(result.sum_sp_values, instance_of(IdValues))
        assert_that(result.sum_sp_values['0'], equal_to(expected))

    @patch_rest
    def test_combine_numeric_values(self):
        result = qr_128.by_path('sp.*.storage.lun.*.writeBlocks')
        other = qr_128.by_path('sp.*.storage.blockSize')
        expected = 4281 * 512 + 67756 * 512
        assert_that(result.combine_numeric_values(other),
                    instance_of(IdValues))
        assert_that(result.combine_numeric_values(other)['sv_677'],
                    equal_to(expected))

    @patch_rest
    def test_combine_sp_values(self):
        result = qr_128.by_path('sp.*.storage.summary.writes')
        other = qr_128.by_path('sp.*.storage.blockSize')
        expected = 1177 * 512
        assert_that(result.combine_sp_values(other), instance_of(IdValues))
        assert_that(result.combine_sp_values(other)['spb'], equal_to(expected))

    @patch_rest
    def test_sum_combined_sp_values(self):
        result = qr_128.by_path('sp.*.storage.summary.writes')
        other = qr_128.by_path('sp.*.storage.blockSize')
        expected = 1597454 * 512 + 1177 * 512
        assert_that(result.sum_combined_sp_values(other),
                    instance_of(IdValues))
        assert_that(result.sum_combined_sp_values(other)['0'],
                    equal_to(expected))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import re

import six

from storops.connection.exceptions import SSHExecutionError
from storops.lib.common import check_int
import storops.vnx.resource.mover
import storops.vnx.resource.nas_pool
from storops.vnx.resource import VNXResource, VNXCliResourceList
from storops.vnx.resource.fs_snap import VNXFsSnap

__author__ = 'Jay Xu'

LOG = logging.getLogger(__name__)


class VNXFileSystemList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXFileSystem

    def _get_raw_resource(self):
        return self._cli.get_filesystem()


class VNXFileSystem(VNXResource):
    def __init__(self, name=None, fs_id=None, cli=None):
        super(VNXFileSystem, self).__init__()
        self._name = name
        self._fs_id = fs_id
        self._cli = cli

    @classmethod
    def get(cls, cli, name=None, fs_id=None):

        if name is not None or fs_id is not None:
            ret = VNXFileSystem(name=name, fs_id=fs_id, cli=cli)
        else:
            ret = VNXFileSystemList(cli=cli)
        return ret

    def get_name(self):
        return self._get_name()

    def get_fs_id(self):
        if self._fs_id is not None:
            ret = self._fs_id
        else:
            ret = self.fs_id
        return ret

    @staticmethod
    def get_id(fs):
        if isinstance(fs, VNXFileSystem):
            fs = fs.get_fs_id()
        try:
            ret = check_int(fs)
        except ValueError:
            raise ValueError('invalid fs id supplied: {}'
                             .format(fs))
        return ret

    def _get_raw_resource(self):
        if self._name is not None:
            ret = self._cli.get_filesystem(self._name)
        elif self._fs_id is not None:
            ret = self._cli.get_filesystem(fs_id=self._fs_id)
        else:
            raise ValueError('fs id or name should be specified.')
        return ret

    @staticmethod
    def create(cli, name, pool, size_kb=None, mover=1, is_vdm=False):
        if size_kb is None:
            # default to 2 MB
            size_kb = 2 * 1024

        pool = storops.vnx.resource.nas_pool.VNXNasPool.get_id(pool)
        mover = storops.vnx.resource.mover.VNXMover.get_id(mover)
        resp = cli.create_filesystem(name, size_kb, pool, mover, is_vdm)
        resp.raise_if_err()
        return VNXFileSystem(name, cli=cli)

    def create_snap(self, name, pool=None):
        if pool is None and self.pools:
            pool = self.pools[0]
        return VNXFsSnap.create(cli=self._cli, name=name, fs=self, pool=pool)

    def delete(self):
        resp = self._cli.delete_filesystem(self.get_fs_id())
        resp.raise_if_err()
        return resp

    def extend(self, new_size, pool=None):
        if pool is None and self.pools:
            pool = self.pools[0]
        pool_id = storops.vnx.resource.nas_pool.VNXNasPool.get_id(pool)
        fs_id = self.get_fs_id()
        delta_size = new_size - self.size
        resp = self._cli.extend_fs(fs_id, delta_size, pool_id)
        resp.raise_if_err()
        return resp

    def create_from_snapshot(self, name, snap_name, source_fs_name, pool_name,
                             mover_name, connect_id):
        # todo: normalize this function and add tests
        create_fs_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
            '-name', name,
            '-type', 'uxfs',
            '-create',
            'samesize=' + source_fs_name,
            'pool=%s' % pool_name,
            'storage=SINGLE',
            'worm=off',
            '-thin', 'no',
            '-option', 'slice=y',
        ]

        self._execute_cmd(create_fs_cmd)

        ro_mount_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_mount', mover_name,
            '-option', 'ro',
            name,
            '/%s' % name,
        ]
        self._execute_cmd(ro_mount_cmd)

        session_name = name + ':' + snap_name
        copy_ckpt_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_copy',
            '-name', session_name[0:63],
            '-source', '-ckpt', snap_name,
            '-destination', '-fs', name,
            '-interconnect',
            'id=%s' % connect_id,
            '-overwrite_destination',
            '-full_copy',
        ]

        try:
            self._execute_cmd(copy_ckpt_cmd, check_exit_code=True)
        except SSHExecutionError as expt:
            message = (("Failed to copy content from snapshot %(snap)s to "
                        "file system %(filesystem)s. Reason: %(err)s.") %
                       {'snap': snap_name,
                        'filesystem': name,
                        'err': six.text_type(expt)})
            LOG.error(message)

        # When an error happens during nas_copy, we need to continue
        # deleting the checkpoint of the target file system if it exists.
        query_fs_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
            '-info', name,
        ]
        out, err = self._execute_cmd(query_fs_cmd)
        re_ckpts = r'ckpts\s*=\s*(.*)\s*'
        m = re.search(re_ckpts, out)
        if m is not None:
            ckpts = m.group(1)
            for ckpt in re.split(',', ckpts):
                umount_ckpt_cmd = [
                    'env', 'NAS_DB=/nas',
                    '/nas/bin/server_umount', mover_name,
                    '-perm', ckpt,
                ]
                self._execute_cmd(umount_ckpt_cmd)
                delete_ckpt_cmd = [
                    'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
                    '-delete', ckpt,
                    '-Force',
                ]
                self._execute_cmd(delete_ckpt_cmd)

        rw_mount_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_mount', mover_name,
            '-option', 'rw',
            name,
            '/%s' % name,
        ]
        self._execute_cmd(rw_mount_cmd)

        filesystem = {
            'name': name,
        }
        self.filesystem_map[name] = self.resource_class(self, filesystem)

        return self.filesystem_map[name]



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import os

from storops.exception import NoIndexException, UnityResourceNotFoundError, \
    UnityNameNotUniqueError, UnityActionNotAllowedError, \
    UnityPerfMonNotEnabledError
from storops.lib.common import clear_instance_cache, instance_cache, \
    get_local_folder
from storops.lib.metric import MetricsDumper
from storops.lib.resource import Resource, ResourceList
from storops.unity import parser
from storops.unity.calculator import calculators
from storops.unity.parser import NestedProperties

__author__ = 'Cedric Zhuang'


class UnityResource(Resource):
    def __init__(self, _id=None, cli=None):
        super(UnityResource, self).__init__()
        self._id = _id
        self._cli = cli
        self._preloaded_properties = None

    @classmethod
    def _get_parser(cls):
        return parser.get_unity_parser(cls.__name__)

    def verify(self):
        if not self.existed:
            raise ValueError(
                'specified {}:{} not exists.'.format(
                    self.__class__.__name__, self.get_id()))
        return self

    def get_id(self):
        ret = None
        if self._id is not None:
            ret = self._id
        elif self._parsed_resource is not None:
            ret = self._parsed_resource.get('id')
            if ret is not None:
                self._id = ret

        if ret is None:
            raise NoIndexException('id is not available for this resource.')
        return ret

    def delete(self, async_mode=False):
        resp = self._cli.delete(self.resource_class, self.get_id(),
                                async_mode=async_mode)
        resp.raise_if_err()
        return resp

    def modify(self, **req_body):
        return self._cli.modify(self.resource_class,
                                self.get_id(), **req_body)

    @property
    def resource_class(self):
        return self._get_parser().name

    @classmethod
    def build_nested_properties_obj(cls):
        return NestedProperties.build(cls.get_nested_properties())

    @classmethod
    def get_nested_properties(cls):
        return None

    def _get_raw_resource(self):
        _id = self.get_id()
        nested_obj = self.build_nested_properties_obj()
        nested_fields = nested_obj.query_fields if nested_obj else None

        res = self._cli.get(self.resource_class, _id,
                            nested_fields=nested_fields)
        # Rest the preloaded the properties to the nested_properties after
        # fetching data from backend
        self.set_preloaded_properties(nested_obj)
        return res

    def _is_updated(self):
        ret = super(UnityResource, self)._is_updated()
        if ret:
            if self.get_preloaded_prop_keys():
                # Return False when only id is parsed besides the
                # preloaded properties
                other = (
                    set(self.parsed_resource.keys()) -
                    set(self.get_preloaded_prop_keys())
                )
                ret = not (
                    len(other) == 1 and
                    len(self.property_names()) -
                    len(self.get_preloaded_prop_keys()) > 1)
            else:
                # Return False when only id is parsed
                ret = not (self.parsed_resource is None or
                           (len(self._parsed_resource) == 1 and
                            len(self.property_names()) > 1))
        return ret

    def _get_properties(self, dec=0):
        if dec < 0 and not self._is_updated():
            props = {'hash': self.__hash__(),
                     'id': self.get_id()}
        else:
            props = super(UnityResource, self)._get_properties(dec)
        return props

    def _parse_raw(self, data):
        return self._get_parser().parse(data, self._preloaded_properties)

    def get_preloaded_prop_keys(self):
        # Returns the preloaded property keys of this object
        # the properties of child object should be skipped

        if not self._preloaded_properties:
            return []
        return self._preloaded_properties.get_properties()

    def _get_property_from_raw(self, item):
        if item in self.metric_names():
            value = self.get_metric_value(item)
        else:
            value = super(UnityResource, self)._get_property_from_raw(item)
            if isinstance(value, UnityResource):
                value.set_cli(self._cli)
        return value

    def property_names(self):
        names = super(UnityResource, self).property_names()
        if self._cli is not None and self._cli.is_perf_metric_enabled(self):
            names.extend(self.metric_names())
        return names

    @property
    def clz_name(self):
        return self._get_parser().resource_class_name

    def metric_names(self):
        return calculators.get_metric_names(self.clz_name)

    def get_metric_value(self, item):
        if not self._cli.is_perf_metric_enabled(self):
            raise UnityPerfMonNotEnabledError()
        return calculators.get_metric_value(
            self.clz_name, item, self._cli, self.get_id())

    def get_metric_timestamp(self):
        curr = self._cli.curr_counter
        if curr is None or len(curr) == 0:
            ret = None
        else:
            ret = curr[0].timestamp
        return ret

    def set_cli(self, cli):
        if cli is not None:
            self._cli = cli

    @classmethod
    def get(cls, cli, _id=None):
        if not isinstance(_id, cls):
            ret = cls(_id=_id, cli=cli)
        else:
            ret = _id
        return ret

    def set_preloaded_properties(self, props):
        self._preloaded_properties = props

    def _get_unity_rsc(self, clz, _id=None, allow_dup=False, **filters):
        ret = clz.get(cli=self._cli, _id=_id, **filters)
        if 'name' in filters and filters['name'] is not None:
            name = filters['name']
            clz_name = clz.get_resource_class().__name__
            if len(ret) == 0:
                raise UnityResourceNotFoundError(
                    '{}:{} not found.'.format(clz_name, name))
            elif len(ret) > 1:
                if not allow_dup:
                    raise UnityNameNotUniqueError(
                        'multiple {} with name {} found.'.format(clz_name,
                                                                 name),
                        # throw out the found multiple objects for later
                        # analysis
                        objects=ret)
            else:
                ret = ret[0]
        return ret

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            ret = (self.get_id() == other.get_id())
        else:
            ret = False
        return ret

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return super(UnityResource, self).__hash__()

    @property
    def doc(self):
        """ Document string of all retrieved properties.

        Return the string that including the description of this resource
        and the the property description of this resource.
        :return: string
        """
        return self._cli.get_doc(self.__class__)

    def action(self, action, **kwargs):
        return self._cli.action(self.resource_class,
                                self.get_id(),
                                action,
                                **kwargs)

    @property
    def system_version(self):
        if self._cli is None:
            return None
        return self._cli.system_version


class UnitySingletonResource(UnityResource):
    def __init__(self, cli=None):
        super(UnitySingletonResource, self).__init__(self.singleton_id(), cli)

    @classmethod
    def singleton_id(cls):
        return '0'

    @classmethod
    def get(cls, cli, _id=None):
        if not isinstance(_id, cls):
            ret = cls(cli=cli)
        else:
            ret = _id
        return ret

    def delete(self, async_mode=False):
        raise UnityActionNotAllowedError()


class UnityAttributeResource(UnityResource):
    """ work as an attributes collection of another resource

    This kind of resource don't have individual get or update methods.
    They work as a collection of the attributes of another resource.
    """

    def _get_raw_resource(self):
        raise '{} is not a independent resource.'.format(
            self.__class__.__name__)


class UnityResourceList(UnityResource, ResourceList):
    def __init__(self, cli=None, **the_filter):
        UnityResource.__init__(self, cli=cli)
        ResourceList.__init__(self)
        self._rsc_filter = the_filter

    @classmethod
    def get_resource_class(cls):
        raise NotImplementedError(
            'resource class for {} not implemented.'.format(cls.__name__))

    @classmethod
    def _get_parser(cls):
        return parser.get_unity_parser(cls.get_resource_class().__name__)

    @clear_instance_cache
    def update(self, data=None):
        ret = super(UnityResourceList, self).update(data)
        for item in self._list:
            item._cli = self._cli
        return ret

    def _get_raw_resource(self):
        the_filter = {}
        _parser = self._get_parser()
        for k, v in self._rsc_filter.items():
            # if k is like host.id for "host.id eq XXX" rest filter
            keys = k.split('.')
            # ingore the left string after '.' since both are ok
            # 'host=<host_id> or {'host.id': <host_id>}'
            label = _parser.get_property_label(keys[0])
            if not label:
                raise ValueError(
                    '"{}" is not a valid property of {}.'.format(
                        k, self.get_resource_class().__name__))
            # support {'host.id': <host_id>}
            if len(keys) == 2:
                label = k
            the_filter[label] = v
        nested_obj = self.get_resource_class().build_nested_properties_obj()
        nested_fields = nested_obj.query_fields if nested_obj else None
        res = self._cli.get_all(
            self.resource_class, the_filter=the_filter,
            nested_fields=nested_fields)
        self.set_preloaded_properties(nested_obj)
        return res

    def set_cli(self, cli):
        super(UnityResourceList, self).set_cli(cli)
        for item in self:
            if isinstance(item, UnityResource):
                item.set_cli(cli)

    @classmethod
    def get(cls, cli, _id=None, **filters):
        if _id is None:
            ret = cls(cli=cli, **filters)
        else:
            ret = cls.get_resource_class().get(cli=cli, _id=_id)
        return ret

    @classmethod
    def get_list(cls, cli, value):
        if value is None:
            ret = None
        elif isinstance(value, cls):
            ret = value
        elif isinstance(value, (tuple, list, set)):
            ret = [cls.get_resource_class().get(cli, v) for v in value]
        else:
            ret = [cls.get_resource_class().get(cli, value)]
        return ret

    @property
    def first_item(self):
        if len(self) > 0:
            ret = self[0]
        else:
            raise ValueError('no instance available for found.')
        return ret

    def _parse_raw(self, data):
        return self._get_parser().parse_all(data)

    def set_preloaded_properties(self, props):
        for i in self:
            i.set_preloaded_properties(props)

    def _get_resource_instance(self):
        return self.get_resource_class()(cli=self._cli)

    def persist_metric_data(self, filename=None):
        if filename is None:
            filename = self.get_default_metric_csv_filename()
        return self._metrics_dumper.persist_metric_data(filename)

    def get_metrics_csv(self, sep=None):
        return self._metrics_dumper.get_metrics_csv(sep=sep)

    def get_default_metric_csv_filename(self):
        folder = get_local_folder()
        name = '{}_{}.csv'.format(self._cli.ip, self.resource_class_name)
        return os.path.join(folder, name)

    @property
    @instance_cache
    def _metrics_dumper(self):
        def hdr_cb(rsc):
            return [str(rsc.get_metric_timestamp()),
                    rsc.get_id(),
                    rsc._get_name()]

        extra_headers = ['timestamp', 'id', 'name']
        return MetricsDumper(self, extra_headers, hdr_cb)

    @property
    def resource_class_name(self):
        return self._get_resource_instance().resource_class

    def _is_updated(self):
        return ResourceList._is_updated(self)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

import ddt
import mock
from hamcrest import equal_to, assert_that, instance_of, raises, none, \
    only_contains, not_none, is_not, calling

from storops.exception import UnityHostIpInUseError, \
    UnityResourceNotFoundError, UnityHostInitiatorNotFoundError, \
    UnityHostInitiatorUnknownType, UnityAluAlreadyAttachedError, \
    UnityAttachAluExceedLimitError, UnitySnapAlreadyPromotedException, \
    SystemAPINotSupported, UnityHostInitiatorExistedError, \
    UnityResourceNotAttachedError, UnityNoHluAvailableError, \
    UnityHluNumberInUseError, UnityAttachExceedLimitError
from storops.unity.enums import HostTypeEnum, HostManageEnum, \
    HostPortTypeEnum, HealthEnum, HostInitiatorTypeEnum, \
    HostInitiatorSourceTypeEnum, HostInitiatorIscsiTypeEnum
from storops.unity.resource.health import UnityHealth
from storops.unity.resource.host import UnityHost, \
    UnityHostInitiator, UnityHostInitiatorList, UnityHostIpPortList, \
    UnityHostList, UnityHostIpPort, UnityHostInitiatorPathList, \
    UnityHostLun, UnityHostLunList, UnityHostInitiatorUpdater
from storops.unity.resource.host_container import UnityHostContainer
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.nfs_share import UnityNfsShare
from storops.unity.resource.snap import UnitySnap
from storops.unity.resource.tenant import UnityTenant
from storops.unity.resource.vmware import UnityDataStoreList, UnityVmList
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


@ddt.ddt
class UnityHostTest(TestCase):
    @patch_rest
    def test_properties(self):
        host = UnityHost(_id='Host_1', cli=t_rest())
        assert_that(host.id, equal_to('Host_1'))
        assert_that(host.type, equal_to(HostTypeEnum.HOST_AUTO))
        assert_that(host.auto_manage_type, equal_to(HostManageEnum.VMWARE))
        assert_that(host.health, instance_of(UnityHealth))
        assert_that(host.name, equal_to('10.244.209.90'))
        assert_that(host.description, equal_to(''))
        assert_that(host.os_type, equal_to('VMware ESXi 6.0.0'))
        assert_that(host.host_pushed_uuid,
                    equal_to('5322a3d1-2901-08c3-c39f-f80f41fafe2e'))
        assert_that(host.host_polled_uuid,
                    equal_to('rfc4122.cd9f4de2-78a5-11e3-85bd-f80f41fafe2e'))
        assert_that(str(host.last_poll_time),
                    equal_to('2016-03-03 04:40:13+00:00'))
        assert_that(host.host_container, instance_of(UnityHostContainer))
        assert_that(host.iscsi_host_initiators,
                    instance_of(UnityHostInitiatorList))
        assert_that(host.host_ip_ports, instance_of(UnityHostIpPortList))
        assert_that(host.datastores, instance_of(UnityDataStoreList))
        assert_that(host.vms, instance_of(UnityVmList))

    @patch_rest
    def test_nested_properties(self):
        host = UnityHost(_id='Host_12', cli=t_rest())
        assert_that(
            host.fc_host_initiators.initiator_id,
            only_contains('20:00:00:00:C9:F3:AB:0C:10:00:00:00:C9:F3:AB:0C',
                          '20:00:00:00:C9:F3:AB:0D:10:00:00:00:C9:F3:AB:0D'))
        assert_that(host.iscsi_host_initiators.initiator_id, only_contains(
            'iqn.1998-01.com.vmware:esx239209-7e7a57a4'))
        assert_that(host.fc_host_initiators[0].paths[0].is_logged_in,
                    equal_to(True))
        assert_that(
            host.fc_host_initiators[1].paths[0].fc_port.wwn,
            equal_to('50:06:01:60:C7:E0:01:DA:50:06:01:6C:47:E0:01:DA'))
        assert_that(
            host.fc_host_initiators[1].paths[0].initiator.type,
            equal_to(HostInitiatorTypeEnum.FC))
        assert_that(host.host_ip_ports[0].address, equal_to('10.245.54.151'))
        assert_that(host.host_luns, instance_of(UnityHostLunList))
        assert_that(host.host_luns[0].lun.name, equal_to('Yangpu'))

    @patch_rest
    def test_nested_properties_shadow_copy(self):
        host = UnityHost(_id='Host_12', cli=t_rest())
        host.fc_host_initiators[0].paths[0].is_logged_in = False
        host.fc_host_initiators[0].paths[0].initiator.type = \
            HostInitiatorTypeEnum.ISCSI
        paths = host.fc_host_initiators[0].paths.shadow_copy()
        assert_that(paths[0].is_logged_in, equal_to(False))
        assert_that(paths[0].initiator.type,
                    equal_to(HostInitiatorTypeEnum.ISCSI))

    @patch_rest
    def test_properties_in_tenant(self):
        host = UnityHost(_id='Host_16', cli=t_rest())
        assert_that(host.id, equal_to('Host_16'))
        assert_that(host.tenant, instance_of(UnityTenant))
        assert_that(host.tenant.id, equal_to('tenant_1'))
        assert_that(host.tenant.vlans, only_contains(1, 3))

    @patch_rest
    def test_get_all(self):
        hosts = UnityHostList(cli=t_rest())
        assert_that(len(hosts), equal_to(7))

    @patch_rest
    def test_get_host_ip_with_mask(self):
        host = UnityHost.get_host(t_rest(), '10.244.209.90/32')
        assert_that(host.ip_list, only_contains('10.244.209.90'))

    @patch_rest
    def test_get_host_in_tenant_with_ip(self):
        host = UnityHost.get_host(t_rest(), '192.168.112.23',
                                  tenant='tenant_1')
        assert_that(host.tenant.id, equal_to('tenant_1'))
        host = UnityHost.get_host(t_rest(), '192.168.112.23')
        assert_that(host.tenant, equal_to(None))

    @patch_rest
    def test_get_host_in_tenant_when_tenant_not_supported(self):
        def do():
            UnityHost.get_host(t_rest('3.1.0'), '192.168.112.23',
                               tenant='tenant_1')

        assert_that(do, raises(SystemAPINotSupported))

    @patch_rest
    def test_get_host_with_force_create(self):
        host = UnityHost.get_host(t_rest(), '192.168.112.24',
                                  tenant='tenant_1',
                                  force_create=True)
        assert_that(host._id, equal_to('Host_15'))

    @patch_rest
    def test_get_host_ipv6_with_mask(self):
        host = UnityHost.get_host(t_rest(), '2001:db8:a0b:12f0::/64',
                                  tenant='tenant_1',
                                  force_create=True)
        assert_that(host, not_none())
        assert_that(host.ip_list, only_contains('2001:db8:a0b:12f0:0:0:0:0'))

    @ddt.data({'version': '3.3.0'},
              {'version': '4.0.1'})
    @ddt.unpack
    @patch_rest
    def test_get_host_with_ip(self, version):
        host = UnityHost.get_host(t_rest(version), '10.244.209.90')
        assert_that(host.ip_list, only_contains('10.244.209.90'))

    @patch_rest
    def test_ip_list_of_host_list(self):
        share = UnityNfsShare(cli=t_rest(), _id='NFSShare_31')
        assert_that(share.read_write_hosts.ip_list, only_contains('1.1.1.1'))
        assert_that(share.read_only_hosts.ip_list, equal_to([]))
        assert_that(share.root_access_hosts, none())

    @patch_rest
    def test_create_simple_host(self):
        host = UnityHost.create(t_rest(), name='host1',
                                host_type=HostTypeEnum.HOST_MANUAL,
                                os='customized os')
        assert_that(host.get_id(), equal_to('Host_11'))
        assert_that(host.name, equal_to('host1'))
        assert_that(host.os_type, equal_to('customized os'))

    @patch_rest
    def test_delete_host_success(self):
        host = UnityHost(cli=t_rest(), _id='Host_11')
        resp = host.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_add_ip_port(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        port = host.add_ip_port('1.1.1.1')
        assert_that(port.existed, equal_to(True))
        assert_that(port.address, equal_to('1.1.1.1'))

    @patch_rest
    def test_delete_ip_port(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        resp = host.delete_ip_port('1.1.1.1')
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_host_and_ip_port(self):
        host = UnityHost(cli=t_rest(), _id='Host_1')
        ip_port = host.host_ip_ports[0]
        resp = host.delete()
        assert_that(resp.is_ok(), equal_to(True))
        assert_that(ip_port.delete, raises(UnityResourceNotFoundError))

    @patch_rest
    def test_create_subset_host(self):
        host = UnityHost.get_host(t_rest(), '7.7.7.7/8', force_create=True)
        assert_that(host.ip_list, only_contains('7.7.7.7'))
        assert_that(host.type, equal_to(HostTypeEnum.SUBNET))

    @patch_rest
    def test_add_initiator(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        wwn = "50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:44:10"
        initiator = host.add_initiator(wwn)
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.existed, equal_to(True))
        assert_that(host.fc_host_initiators,
                    instance_of(UnityHostInitiatorList))

    @patch_rest
    def test_add_initiator_iscsi(self):
        host = UnityHost(cli=t_rest(), _id='Host_1')
        iqn = "iqn.1993-08.org.debian:01:a4f95ed19999"
        initiator = host.add_initiator(iqn)
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.existed, equal_to(True))
        assert_that(host.iscsi_host_initiators,
                    instance_of(UnityHostInitiatorList))

    @patch_rest
    def test_add_initiator_iscsi_already_existed(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        iqn = "iqn.1993-08.org.debian:01:a4f95ed19999"

        def _inner():
            host.add_initiator(iqn)

        assert_that(_inner, raises(UnityHostInitiatorExistedError))

    @patch_rest
    def test_add_not_exist_initiator_with_force(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        wwn = "50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:99:99"
        initiator = host.add_initiator(wwn, force_create=True)
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.existed, equal_to(True))
        assert_that(host.fc_host_initiators,
                    instance_of(UnityHostInitiatorList))

    @patch_rest
    def test_add_initiator_not_exist(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        wwn = "50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:99:99"

        def f():
            host.add_initiator(wwn, force_create=False)

        assert_that(f, raises(UnityHostInitiatorNotFoundError))

    @patch_rest
    def test_delete_initiator(self):
        host = UnityHost(cli=t_rest(), _id='Host_1')
        wwn = "50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:44:10"
        resp = host.delete_initiator(wwn)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_initiator_not_found(self):
        host = UnityHost(cli=t_rest(), _id='Host_1')
        wwn = "50:00:14:40:47:B0:0C:44:50:00:14:42:99:99:99:99"

        def f():
            host.delete_initiator(wwn)

        assert_that(f, raises(UnityHostInitiatorNotFoundError))

    @patch_rest
    def test_host_lun(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        assert_that(host.host_luns, instance_of(UnityHostLunList))
        assert_that(len(host.host_luns), equal_to(2))

    @patch_rest
    def test_get_all_host_lun_all(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        all_luns = host._get_host_luns()
        assert_that(len(all_luns), equal_to(2))

    @patch_rest
    def test_get_one_host_lun(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun1 = UnityLun(cli=t_rest(), _id="sv_2")
        which = host._get_host_luns(lun1)
        assert_that(len(which), equal_to(1))
        assert_that(which[0].lun.id, equal_to(lun1.id))

    @patch_rest
    def test_has_alu_true(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun1 = UnityLun(cli=t_rest(), _id="sv_2")
        has = host.has_alu(lun1)
        assert_that(has, equal_to(True))

    @patch_rest
    def test_has_alu_false(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun2 = UnityLun(cli=t_rest(), _id="sv_4")
        has = host.has_alu(lun2)
        assert_that(has, equal_to(False))

    @patch_rest
    def test_get_hlu(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_2")
        # UnityResourceList will return the found UnityResource
        # When '_id' as filter.
        host_lun = UnityHostLunList.get(cli=t_rest(), _id="Host_10_sv_2_prod")
        assert_that(host_lun, instance_of(UnityHostLun))
        hlu = host.get_hlu(lun)
        assert_that(hlu, equal_to(host_lun.hlu))

        # Be caureful, this will return UnityResourceList when 'id' as filter
        host_lun2 = UnityHostLunList.get(cli=t_rest(), id="Host_10_sv_2_prod")
        assert_that(host_lun2, instance_of(UnityHostLunList))
        assert_that(len(host_lun2), equal_to(1))
        assert_that(host_lun, equal_to(host_lun2[0]))

    @patch_rest
    def test_get_hlu_using_object_id_filter(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_2")
        filters = {'host.id': host.id, 'lun.id': lun.id}
        host_lun = UnityHostLunList.get(cli=t_rest(), **filters)
        assert_that(host_lun, instance_of(UnityHostLunList))
        assert_that(len(host_lun), equal_to(1))
        hlu = host.get_hlu(lun)
        assert_that(hlu, equal_to(host_lun[0].hlu))

    @patch_rest
    def test_get_hlu_not_found(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_4")
        hlu = host.get_hlu(lun)
        assert_that(hlu, equal_to(None))

    @patch_rest
    def test_detach_alu_without_host_access(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_2")
        resp = host.detach_alu(lun)
        assert_that(resp, equal_to(None))

    @patch_rest
    def test_detach_attached_hlu(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_2")
        resp = host.detach_alu(lun)
        assert_that(resp, equal_to(None))

    @patch_rest
    def test_detach_alu(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_4")
        resp = host.detach_alu(lun)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_attach_attached_hlu(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_2")

        def f():
            host.attach_alu(lun)

        assert_that(f, raises(UnityAluAlreadyAttachedError))

    @patch_rest
    def test_attach_alu(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_4")
        hlu = host.attach_alu(lun)
        assert_that(hlu, equal_to(None))

    @patch_rest
    def test_attach_alu_exceed_limit(self):
        host = UnityHost(cli=t_rest(), _id='Host_11')
        lun = UnityLun(cli=t_rest(), _id="sv_2")

        def f():
            host.attach_alu(lun)

        assert_that(f, raises(UnityAttachAluExceedLimitError))

    @patch_rest
    def test_random_hlu_number(self):
        host = UnityHost(cli=t_rest(), _id='Host_23')
        assert_that(host._random_hlu_number(), is_not(equal_to(0)))

    @patch_rest
    def test_random_hlu_number_no_available(self):
        host = UnityHost(cli=t_rest(), _id='Host_23')
        with mock.patch('storops.unity.resource.host.MAX_HLU_NUMBER', 0):
            assert_that(calling(host._random_hlu_number),
                        raises(UnityNoHluAvailableError))

    @patch_rest
    def test_attach_with_retry_skip_hlu_0(self):
        host = UnityHost(cli=t_rest(), _id='Host_23')
        lun = UnityLun(_id='sv_5610', cli=t_rest())
        lun.is_cg_member = False
        with mock.patch('storops.unity.resource.host.UnityHost.'
                        '_random_hlu_number', new=lambda _: 12781):
            hlu = host._attach_with_retry(lun, True)
            assert_that(hlu, is_not(equal_to(0)))
            assert_that(hlu, equal_to(12781))

    @patch_rest
    def test_attach_with_retry_member_snap_skip_hlu_0(self):
        host = UnityHost(cli=t_rest(), _id='Host_27')
        snap = UnitySnap(_id='38654705831', cli=t_rest())
        with mock.patch('storops.unity.resource.host.UnityHost.'
                        '_random_hlu_number', new=lambda _: 12781):
            hlu = host._attach_with_retry(snap, True)
            assert_that(hlu, is_not(equal_to(0)))
            assert_that(hlu, equal_to(12781))

    @patch_rest
    def test_attach_with_retry_no_skip_hlu_0(self):
        host = UnityHost(cli=t_rest(), _id='Host_23')
        lun = UnityLun(_id='sv_5610', cli=t_rest())
        lun.is_cg_member = False
        assert_that(host._attach_with_retry(lun, False), equal_to(0))

    @patch_rest
    def test_attach_with_retry_no_retry(self):
        host = UnityHost(cli=t_rest(), _id='Host_24')
        lun = UnityLun(_id='sv_5610', cli=t_rest())
        lun.is_cg_member = False
        assert_that(host._attach_with_retry(lun, True), equal_to(1))

    @patch_rest
    def test_attach_with_retry_hlu_in_use(self):
        host = UnityHost(cli=t_rest(), _id='Host_23')
        lun = UnityLun(_id='sv_5610', cli=t_rest())
        lun.is_cg_member = False
        with mock.patch('storops.unity.resource.host.UnityHost.'
                        '_modify_hlu',
                        new=mock.Mock(side_effect=UnityHluNumberInUseError)):
            assert_that(calling(host._attach_with_retry).with_args(lun, True),
                        raises(UnityHluNumberInUseError))

    @patch_rest
    def test_attach_with_retry_hlu_in_use_but_no_retry(self):
        host = UnityHost(cli=t_rest(), _id='Host_24')
        lun = UnityLun(_id='sv_5610', cli=t_rest())
        lun.is_cg_member = False
        with mock.patch('storops.unity.resource.host.UnityHost.'
                        '_modify_hlu',
                        new=mock.Mock(side_effect=UnityHluNumberInUseError)):
            assert_that(host._attach_with_retry(lun, True), equal_to(1))

    @patch_rest
    def test_attach_with_retry_no_hlu_available(self):
        host = UnityHost(cli=t_rest(), _id='Host_23')
        lun = UnityLun(_id='sv_5610', cli=t_rest())
        lun.is_cg_member = False
        with mock.patch('storops.unity.resource.host.UnityHost.'
                        '_random_hlu_number',
                        new=mock.Mock(side_effect=UnityNoHluAvailableError)):
            assert_that(calling(host._attach_with_retry).with_args(lun, True),
                        raises(UnityNoHluAvailableError))

    @patch_rest
    def test_attach_exceptions(self):
        host = UnityHost(cli=t_rest(), _id='Host_23')
        lun = UnityLun(_id='sv_5609', cli=t_rest())
        with mock.patch('storops.unity.resource.host.UnityHost.'
                        '_attach_with_retry',
                        new=mock.Mock(side_effect=SystemAPINotSupported)):
            assert_that(calling(host.attach).with_args(lun, skip_hlu_0=True),
                        raises(SystemAPINotSupported))
        with mock.patch(
                'storops.unity.resource.host.UnityHost._attach_with_retry',
                new=mock.Mock(side_effect=UnityAttachExceedLimitError)):
            assert_that(calling(host.attach).with_args(lun, skip_hlu_0=True),
                        raises(UnityAttachExceedLimitError))

    @patch_rest
    def test_attach_exceptions_detach(self):
        host = UnityHost(cli=t_rest(), _id='Host_25')
        lun = UnityLun(_id='sv_5611', cli=t_rest())
        lun.is_cg_member = False
        with mock.patch('storops.unity.resource.host.UnityHost.'
                        '_attach_with_retry',
                        new=mock.Mock(side_effect=UnityHluNumberInUseError)):
            assert_that(calling(host.attach).with_args(lun, skip_hlu_0=True),
                        raises(UnityHluNumberInUseError))
        with mock.patch(
                'storops.unity.resource.host.UnityHost._attach_with_retry',
                new=mock.Mock(side_effect=UnityNoHluAvailableError)):
            assert_that(calling(host.attach).with_args(lun, skip_hlu_0=True),
                        raises(UnityNoHluAvailableError))

    @patch_rest
    def test_attach_exceptions_detach_dummy_lun(self):
        host = UnityHost(cli=t_rest(), _id='Host_26')
        lun = UnityLun(_id='sv_5611', cli=t_rest())
        lun.is_cg_member = False
        with mock.patch('storops.unity.resource.host.UnityHost.'
                        '_attach_with_retry',
                        new=mock.Mock(side_effect=UnityHluNumberInUseError)):
            assert_that(calling(host.attach).with_args(lun, skip_hlu_0=True),
                        raises(UnityHluNumberInUseError))
        with mock.patch(
                'storops.unity.resource.host.UnityHost._attach_with_retry',
                new=mock.Mock(side_effect=UnityNoHluAvailableError)):
            assert_that(calling(host.attach).with_args(lun, skip_hlu_0=True),
                        raises(UnityNoHluAvailableError))

    @patch_rest
    def test_attach_snap_skip_first_hlu(self):
        host = UnityHost(cli=t_rest(), _id='Host_11')
        snap = UnitySnap(_id='38654705676', cli=t_rest())
        assert_that(calling(host.attach).with_args(snap, skip_hlu_0=True),
                    raises(UnitySnapAlreadyPromotedException))

    @patch_rest
    def test_get_attached_cg_snap_hlu(self):
        host = UnityHost(cli=t_rest(), _id='Host_22')
        snap = UnitySnap(cli=t_rest(), _id='85899345930')
        lun = UnityLun(cli=t_rest(), _id='sv_3338')
        assert_that(host.get_hlu(snap, lun), equal_to(2))

    @patch_rest
    def test_get_hlu_of_cg_member(self):
        host = UnityHost(cli=t_rest(), _id='Host_22')
        lun = UnityLun(cli=t_rest(), _id='sv_3338')
        assert_that(host.get_hlu(lun), equal_to(3))

    @patch_rest
    def test_host_modify(self):
        host = UnityHost(cli=t_rest(), _id='Host_22')
        new_host = host.modify(name="new_host", desc="new Desc", os="Ubuntu")
        assert_that(host, equal_to(new_host))

    @patch_rest
    def test_update_initators(self):
        host = UnityHost(cli=t_rest(), _id='Host_22')
        r = host.update_initiators(
            iqns=["iqn.1993-08.org.debian:01:a4f95ed19999"],
            wwns=["50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:44:10"])
        assert_that(r, equal_to(3))

    @patch_rest
    def test_update_initators_with_fc(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        r = host.update_initiators(
            wwns=["50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:44:11",
                  "50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:44:12"])
        assert_that(r, equal_to(3))

    @patch_rest
    def test_modify_host_lun_not_attached(self):
        host = UnityHost(cli=t_rest(), _id='Host_19')
        lun = UnityLun(cli=t_rest(), _id='sv_3338')
        assert_that(calling(host.modify_host_lun).with_args(lun, 100),
                    raises(UnityResourceNotAttachedError))

    @patch_rest
    def test_modify_host_lun_hlu_in_use(self):
        host = UnityHost(cli=t_rest(), _id='Host_22')
        lun = UnityLun(cli=t_rest(), _id='sv_3338')
        assert_that(calling(host.modify_host_lun).with_args(lun, 3),
                    raises(UnityHluNumberInUseError))

    @patch_rest
    def test_modify_host_lun(self):
        host = UnityHost(cli=t_rest(), _id='Host_22')
        lun = UnityLun(cli=t_rest(), _id='sv_3338')
        host.modify_host_lun(lun, 100)

    @patch_rest
    def test_has_hlu(self):
        host = UnityHost(cli=t_rest(), _id='Host_10')
        lun = UnityLun(cli=t_rest(), _id="sv_2")
        assert_that(True, equal_to(host.has_hlu(lun)))

    @patch_rest
    def test_has_hlu_false(self):
        host = UnityHost(cli=t_rest(), _id='Host_11')
        lun = UnityLun(cli=t_rest(), _id="sv_2")
        assert_that(False, equal_to(host.has_hlu(lun)))

    @patch_rest
    def test_has_hlu_snap(self):
        host = UnityHost(cli=t_rest(), _id='Host_27')
        snap = UnitySnap(cli=t_rest(), _id="38654705831")
        assert_that(True, equal_to(host.has_hlu(snap)))

    @patch_rest
    def test_has_hlu_snap_cg_member(self):
        host = UnityHost(cli=t_rest(), _id='Host_27')
        cg_snap = UnitySnap(cli=t_rest(), _id="85899345958")
        lun_in_cg = UnityLun(cli=t_rest(), _id='sv_58')
        assert_that(True, equal_to(host.has_hlu(cg_snap, cg_member=lun_in_cg)))


class UnityHostInitiatorUpdaterTest(TestCase):
    def test_compute(self):
        updater = UnityHostInitiatorUpdater(
            "fake_host",
            {'iqn1', 'iqn2', 'wwn1', 'wwn3'},
            {'iqn3', 'iqn1', 'wwn1'})
        to_add, to_delete = updater.compute()
        assert_that(to_add, equal_to({'iqn3'}))
        assert_that(to_delete, equal_to({'iqn2', 'wwn3'}))


class UnityHostIpPortTest(TestCase):
    @patch_rest
    def test_properties(self):
        ip_port = UnityHostIpPort(cli=t_rest(), _id='HostNetworkAddress_1')
        assert_that(ip_port.host, instance_of(UnityHost))
        assert_that(ip_port.address, equal_to('10.244.209.90'))
        assert_that(ip_port.type, equal_to(HostPortTypeEnum.IPv4))

    @patch_rest
    def test_get_all(self):
        ip_ports = UnityHostIpPortList(cli=t_rest())
        assert_that(len(ip_ports), equal_to(8))

    @patch_rest
    def test_create_success(self):
        ip_port = UnityHostIpPort.create(t_rest(), 'Host_9', '1.1.1.1')
        assert_that(ip_port.address, equal_to('1.1.1.1'))
        assert_that(ip_port.type, equal_to(HostPortTypeEnum.IPv4))
        assert_that(ip_port.existed, equal_to(True))

    @patch_rest
    def test_create_ip_in_use(self):
        def f():
            UnityHostIpPort.create(t_rest(), 'Host_1', '1.1.1.1')

        assert_that(f, raises(UnityHostIpInUseError, 'already exists'))

    @patch_rest
    def test_delete_success(self):
        ip_port = UnityHostIpPort(cli=t_rest(), _id='HostNetworkAddress_10')
        resp = ip_port.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_get_by_ip(self):
        ip_ports = UnityHostIpPortList(cli=t_rest(), address='10.244.209.90')
        assert_that(len(ip_ports), equal_to(1))
        assert_that(ip_ports[0].address, equal_to('10.244.209.90'))


class UnityHostInitiatorTest(TestCase):
    @patch_rest
    def test_fc_initiator_properties(self):
        initiator = UnityHostInitiator(cli=t_rest(), _id='HostInitiator_2')
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.existed, equal_to(True))
        assert_that(initiator.health, instance_of(UnityHealth))
        assert_that(initiator.health.value, equal_to(HealthEnum.OK))
        assert_that(initiator.type, equal_to(HostInitiatorTypeEnum.FC))
        wwn = "50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:44:10"
        assert_that(initiator.initiator_id, equal_to(wwn))
        assert_that(initiator.parent_host, instance_of(UnityHost))
        assert_that(initiator.is_ignored, equal_to(False))
        assert_that(initiator.is_chap_secret_enabled, equal_to(False))
        assert_that(initiator.node_wwn, equal_to("50:00:14:40:47:B0:0C:44"))
        assert_that(initiator.port_wwn, equal_to("50:00:14:42:D0:0C:44:10"))
        assert_that(initiator.paths, instance_of(UnityHostInitiatorPathList))
        assert_that(initiator.source_type,
                    equal_to(HostInitiatorSourceTypeEnum.OPEN_NATIVE))

    @patch_rest
    def test_iscsi_initiator_properties(self):
        initiator = UnityHostInitiator(cli=t_rest(), _id='HostInitiator_3')
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.existed, equal_to(True))
        assert_that(initiator.health, instance_of(UnityHealth))
        assert_that(initiator.health.value, equal_to(HealthEnum.OK))
        assert_that(initiator.type, equal_to(HostInitiatorTypeEnum.ISCSI))
        iqn = "iqn.1993-08.org.debian:01:a4f95ed14d65"
        assert_that(initiator.initiator_id, equal_to(iqn))
        assert_that(initiator.parent_host, instance_of(UnityHost))
        assert_that(initiator.is_ignored, equal_to(False))
        assert_that(initiator.is_chap_secret_enabled, equal_to(True))
        assert_that(initiator.paths, instance_of(UnityHostInitiatorPathList))
        assert_that(initiator.chap_user_name, equal_to(iqn))
        assert_that(initiator.iscsi_type,
                    equal_to(HostInitiatorIscsiTypeEnum.SOFTWARE))
        assert_that(initiator.is_bound, equal_to(False))
        assert_that(initiator.source_type,
                    equal_to(HostInitiatorSourceTypeEnum.DELL))

    @patch_rest
    def test_fc_initiator_create(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        type = HostInitiatorTypeEnum.FC
        wwn = "50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:44:10"
        initiator = UnityHostInitiator.create(t_rest(), wwn, host, type)
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.initiator_id, equal_to(wwn))

    @patch_rest
    def test_iscsi_initiator_create(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        type = HostInitiatorTypeEnum.ISCSI
        iqn = "iqn.1993-08.org.debian:01:a4f95ed14d65"
        initiator = UnityHostInitiator.create(t_rest(), iqn, host, type)
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.initiator_id, equal_to(iqn))

    @patch_rest
    def test_unknown_initiator_create(self):
        host = UnityHost(cli=t_rest(), _id='Host_9')
        type = HostInitiatorTypeEnum.UNKNOWN
        iqn = "iqn.1993-08.org.debian:01:a4f95ed14d65"

        def f():
            UnityHostInitiator.create(t_rest(), iqn, host, type)

        assert_that(f, raises(UnityHostInitiatorUnknownType))

    @patch_rest
    def test_initiator_modify(self):
        initiator = UnityHostInitiator(cli=t_rest(), _id='HostInitiator_2')
        assert_that(initiator.parent_host, instance_of(UnityHost))
        assert_that(initiator.parent_host.id, equal_to("Host_1"))
        host = UnityHost(cli=t_rest(), _id='Host_19')
        resp = initiator.modify(host=host)
        assert_that(resp.is_ok(), equal_to(True))
        assert_that(initiator.parent_host, instance_of(UnityHost))

    @patch_rest
    def test_initiator_delete(self):
        initiator = UnityHostInitiator(cli=t_rest(), _id='HostInitiator_2')
        resp = initiator.delete()
        assert_that(resp.is_ok(), equal_to(True))


class UnityHostInitiatorPathListTest(TestCase):
    @patch_rest
    def test_filter(self):
        is_logged_in = True

        all_paths = UnityHostInitiatorPathList(cli=t_rest())
        paths = all_paths.shadow_copy(is_logged_in=is_logged_in)
        assert_that(len(paths), equal_to(2))
        assert_that(all(p.is_logged_in for p in paths), equal_to(True))

        paths = UnityHostInitiatorPathList(cli=t_rest(),
                                           is_logged_in=is_logged_in)
        assert_that(len(paths), equal_to(2))
        assert_that(all(p.is_logged_in for p in paths), equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.exception import VNXCreateConsistencyGroupError

from storops.vnx.resource.snap import VNXSnap

import storops.vnx.resource.lun
from storops.vnx.resource import VNXCliResource, VNXCliResourceList
from storops import exception as ex

__author__ = 'Cedric Zhuang'


class VNXConsistencyGroupList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXConsistencyGroup

    def _get_raw_resource(self):
        return self._cli.get_cg(poll=self.poll)

    def delete_member(self, *lun_list):
        for cg in self:
            cg.delete_member(*lun_list)


class VNXConsistencyGroup(VNXCliResource):
    def __init__(self, name=None, cli=None):
        super(VNXConsistencyGroup, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        return self._cli.get_cg(name=self._name, poll=self.poll)

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXConsistencyGroupList(cli)
        else:
            ret = VNXConsistencyGroup(name, cli)
        return ret

    @classmethod
    def create(cls, cli, name, members=None, auto_delete=None):
        out = cli.create_cg(name, members, auto_delete)
        ex.raise_if_err(out, 'error creating cg {}.'.format(name),
                        default=VNXCreateConsistencyGroupError)
        return VNXConsistencyGroup(name=name, cli=cli)

    def delete(self):
        name = self._get_name()
        out = self._cli.delete_cg(name, poll=self.poll)
        ex.raise_if_err(out, 'error remove cg "{}".'.format(name),
                        default=ex.VNXConsistencyGroupError)

    def _cg_member_op(self, op, lun_list):
        clz = storops.vnx.resource.lun.VNXLun
        id_list = clz.get_id_list(*lun_list)
        name = self._get_name()
        out = op(name, *id_list, poll=self.poll)
        ex.raise_if_err(out, 'error change member of "{}".'.format(name),
                        default=ex.VNXConsistencyGroupError)

    def add_member(self, *lun_list):
        self._cg_member_op(self._cli.add_cg_member, lun_list)

    def delete_member(self, *lun_list):
        lun_list = list(filter(self.has_member, lun_list))
        self._cg_member_op(self._cli.delete_cg_member, lun_list)

    def replace_member(self, *lun_list):
        self._cg_member_op(self._cli.replace_cg_member, lun_list)

    def has_member(self, lun):
        clz = storops.vnx.resource.lun.VNXLun
        try:
            lun_id = clz.get_id(lun)
        except ValueError:
            # lun id not valid, lun not exists
            lun_id = None

        return (self.lun_list and
                lun_id is not None and
                lun_id in self.lun_list.lun_id)

    def create_snap(self, name, allow_rw=None, auto_delete=None):
        return VNXSnap.create(self._cli, self._get_name(), name, allow_rw,
                              auto_delete)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, raises, equal_to, has_items, \
    only_contains, none, close_to, not_none, greater_than
from storops.vnx.enums import VNXPoolRaidType

from storops_test.vnx.cli_mock import t_cli, patch_cli
from storops_test.vnx.resource.test_lun import get_lun_list
from storops_test.vnx.resource.verifiers import verify_pool_0
from storops.exception import VNXCreatePoolError, VNXPoolNotFoundError, \
    VNXDiskUsedError, VNXPoolNameInUseError, VNXPoolDestroyingError, \
    VNXNameInUseError
from storops.vnx.resource.block_pool import VNXPool, VNXPoolList, \
    VNXPoolFeature

__author__ = 'Cedric Zhuang'


class VNXPoolTest(TestCase):
    @staticmethod
    def get_pool_with_id(pool_id=0):
        return VNXPool(pool_id=pool_id, cli=t_cli())

    @staticmethod
    def get_pool_with_name(name='Pool4File'):
        return VNXPool(name=name, cli=t_cli())

    @patch_cli
    def test_property_not_exist(self):
        def f():
            pool = VNXPool(pool_id=0)
            getattr(pool, '_abc')

        assert_that(f, raises(AttributeError))

    def test_get_pool_id_from_self(self):
        assert_that(self.get_pool_with_id(12).get_pool_id(), equal_to(12))

    @patch_cli
    def test_get_pool_id_from_cli(self):
        assert_that(self.get_pool_with_name().get_pool_id(), equal_to(0))

    @patch_cli
    def test_pool_by_id(self):
        pool = self.get_pool_with_id()
        verify_pool_0(pool)

    @patch_cli
    def test_get_pool_get_all(self):
        pools = VNXPool.get(t_cli())
        assert_that(len(pools), equal_to(5))

    @patch_cli
    def test_get_pool_from_list(self):
        pools = VNXPool.get(t_cli())
        pool = next(p for p in pools if p.pool_id == 4)
        # no error should be thrown here
        pool.update()

    @patch_cli
    def test_get_pool_get_by_name(self):
        pool = VNXPool.get(t_cli(), name='Pool4File')
        verify_pool_0(pool)

    @patch_cli
    def test_get_pool_get_by_id(self):
        pool = VNXPool.get(t_cli(), pool_id=0)
        verify_pool_0(pool)

    @patch_cli
    def test_pool_by_name(self):
        pool = self.get_pool_with_name()
        verify_pool_0(pool)

    @patch_cli
    def test_get_lun(self):
        pool = VNXPool(pool_id=1, cli=t_cli())
        assert_that(pool.name, equal_to('Pool_daq'))
        lun_list = pool.get_lun()
        assert_that(len(lun_list), equal_to(50))
        assert_that(pool.lun_list, equal_to(lun_list))
        assert_that(len(set(lun_list.pool_name)), equal_to(1))

    @patch_cli
    def test_get_disk(self):
        pool = VNXPool(pool_id=1, cli=t_cli())
        disks = pool.disks
        assert_that(len(disks), equal_to(3))
        assert_that(disks.serial_number,
                    only_contains('6XS2EAKG', 'S0PFNECC304969', '6XS2QCG1'))

    @patch_cli
    def test_rename_name_existed(self):
        def f():
            pool = VNXPool(name='p1', cli=t_cli())
            pool.name = 'p2'

        assert_that(f, raises(VNXNameInUseError, 'Name'))

    @patch_cli
    def test_create_pool_success(self):
        def f():
            VNXPool.create(t_cli(), 'p0', ['1_0_0', '1_0_1'])

        assert_that(f, raises(VNXCreatePoolError,
                              'less than minimum required'))

    @patch_cli
    def test_create_pool_name_in_use_0(self):
        def f():
            VNXPool.create(t_cli(), 'p0', ['1_0_0', '1_0_1'], 'r_6')

        assert_that(f, raises(VNXPoolNameInUseError, 'already in use'))

    @patch_cli
    def test_create_pool_name_in_use_1(self):
        def f():
            VNXPool.create(t_cli(), 'p0', ['1_0_0', '1_0_1'], 'r_0')

        assert_that(f, raises(VNXPoolNameInUseError, 'name is already used'))

    @patch_cli
    def test_create_pool_disk_used(self):
        def f():
            VNXPool.create(t_cli(), 'p0', ['1_0_0', '1_0_1'], 'r_10')

        assert_that(f, raises(VNXDiskUsedError, 'already part of'))

    @patch_cli
    def test_create_pool_invalid_disk_number(self):
        def f():
            VNXPool.create(t_cli(), 'p0', ['1_0_0', '1_0_1'], 'r_5')

        assert_that(f, raises(VNXCreatePoolError, 'multiple of 5'))

    @patch_cli
    def test_delete_pool_not_found(self):
        def f():
            VNXPool(0, cli=t_cli()).delete()

        assert_that(f, raises(VNXPoolNotFoundError, 'may not exist'))

    @patch_cli
    def test_delete_pool_destroying(self):
        def f():
            VNXPool(1, cli=t_cli()).delete()

        assert_that(f, raises(VNXPoolDestroyingError, 'is Destroying'))

    @patch_cli
    def test_force_delete_pool(self):
        def f():
            VNXPool(0, cli=t_cli()).delete(True)

        assert_that(f, raises(VNXPoolNotFoundError, 'may not exist'))

    @patch_cli
    def test_update_with_one_key_only(self):
        pool = VNXPool(0, 'p0', t_cli())
        assert_that(pool.consumed_capacity_gbs, equal_to(540.303))

    @patch_cli
    def test_create_lun_ignore_threshold(self):
        def f():
            pool = VNXPool(1, cli=t_cli())
            pool.create_lun('abc', ignore_thresholds=True)

        assert_that(f, raises(VNXPoolNotFoundError, 'may not exist'))

    @patch_cli
    def test_create_lun_with_pool_name(self):
        def f():
            pool = VNXPool(name='p0', cli=t_cli())
            pool.create_lun(lun_id=12)

        assert_that(f, raises(VNXPoolNotFoundError, 'may not exist'))

    @property
    @patch_cli
    def pool(self):
        lun_list = get_lun_list()
        return VNXPool(pool_id=0, cli=t_cli(), system_lun_list=lun_list)

    @patch_cli
    def test_shadow_copy_lun_list(self):
        lun_list = get_lun_list()
        assert_that(self.pool.lun_list.timestamp, equal_to(lun_list.timestamp))

    @patch_cli
    def test_pool_read_iops(self):
        assert_that(self.pool.read_iops, equal_to(3.5))

    @patch_cli
    def test_pool_write_iops(self):
        assert_that(self.pool.write_iops, equal_to(9.0))

    @patch_cli
    def test_pool_total_iops(self):
        assert_that(self.pool.total_iops, equal_to(12.5))

    @patch_cli
    def test_pool_read_mbps(self):
        assert_that(self.pool.read_mbps, equal_to(2.3 + 4.6))

    @patch_cli
    def test_pool_write_mbps(self):
        assert_that(self.pool.write_mbps, equal_to(2.7 + 5.4))

    @patch_cli
    def test_pool_total_mbps(self):
        assert_that(self.pool.total_mbps, equal_to(5.0 + 10.0))

    @patch_cli
    def test_pool_read_size_kb(self):
        assert_that(self.pool.read_size_kb, close_to(2018, 1))

    @patch_cli
    def test_sg_write_size_kb(self):
        assert_that(self.pool.write_size_kb, close_to(921, 1))


class VNXPoolListTest(TestCase):
    @staticmethod
    def get_pool_list():
        return VNXPoolList(t_cli())

    @patch_cli
    def test_pool_list(self):
        pools = self.get_pool_list()
        assert_that(len(pools), equal_to(5))

    @patch_cli
    def test_update(self):
        pools = self.get_pool_list()
        pools.update()
        # call twice
        pools.update()
        assert_that(len(pools), equal_to(5))

    @patch_cli
    def test_get_with_shadow_copy(self):
        lun_list = get_lun_list()
        pools = VNXPool.get(t_cli(), system_lun_list=lun_list)
        assert_that(lun_list.timestamp, not_none())
        assert_that(len(pools), greater_than(0))
        for pool in pools:
            assert_that(pool.lun_list.timestamp, equal_to(lun_list.timestamp))


class VNXPoolTierTest(TestCase):
    @patch_cli
    def test_pool_tier(self):
        pool1 = VNXPool.get(t_cli(), name='tier_test')
        assert_that(pool1.tiers, not_none())
        assert_that(len(pool1.tiers), equal_to(2))
        tier = pool1.tiers[0]

        assert_that(tier.tier_name, equal_to('Performance'))
        assert_that(tier.raid_type, equal_to(VNXPoolRaidType.RAID5))
        assert_that(tier.user_capacity_gb, equal_to(1639.70))
        assert_that(tier.consumed_capacity_gb, equal_to(129.25))
        assert_that(tier.available_capacity_gb, equal_to(1510.44))
        assert_that(tier.percent_subscribed, equal_to(7.88))
        assert_that(tier.data_targeted_for_higher_tier_gb, equal_to(0.0))
        assert_that(tier.data_targeted_for_lower_tier, equal_to(0.0))
        assert_that(tier.data_targeted_within_tier, equal_to(0.0))


class VNXPoolFeatureTest(TestCase):
    @patch_cli
    def test_properties(self):
        f = VNXPoolFeature(t_cli())
        assert_that(f.is_virtual_provisioning_supported, equal_to(True))
        assert_that(f.max_pools, equal_to(20))
        assert_that(f.max_disks_per_pool, equal_to(121))
        assert_that(f.max_disks_for_all_pools, equal_to(121))
        assert_that(f.max_disks_per_operation, equal_to(40))
        assert_that(f.max_pool_luns, equal_to(512))
        assert_that(f.min_pool_lun_size_blocks, equal_to(1))
        assert_that(f.max_pool_lun_size_blocks, equal_to(34359738368))
        assert_that(f.max_pool_lun_size_gbs, equal_to(16384.0))
        assert_that(f.total_number_of_pools, equal_to(4))
        assert_that(f.total_pool_luns, equal_to(351))
        assert_that(f.total_thin_luns, equal_to(3))
        assert_that(f.total_non_thin_luns, none())
        assert_that(f.number_of_disks_used_in_pools, equal_to(18))
        assert_that(f.available_disks.index,
                    has_items('0_0_B8', '0_0_B9'))
        assert_that(f.background_operation_state, none())
        assert_that(f.background_rate, none())

    @patch_cli
    def test_available_disks(self):
        f = VNXPoolFeature(t_cli())
        assert_that(len(f.available_disks), equal_to(5))
